from functools import partial

from validator import decorator
from validator import metadata_helpers


@decorator.register_test(tier=1)
def test_package_json_params(err, xpi_package=None):
    if err.get_resource('has_package_json'):
        validate_required_field(err, 'id', validate_id)
        validate_required_field(err, 'name', validate_name)
        validate_required_field(err, 'version', validate_version)


def validate_required_field(err, field, validate_fn):
    package_json = err.get_resource('package_json')
    if field in package_json:
        value = package_json[field]
        validate_fn(err, value)
    else:
        err.error(
            ('package_json', 'field_required', field),
            'Your package.json is missing a required field',
            'Your package.json is missing the "{field}" field.'.format(
                field=field))


validate_id = partial(metadata_helpers.validate_id, source='package.json')
validate_version = partial(
    metadata_helpers.validate_version, source='package.json')
validate_name = partial(metadata_helpers.validate_name, source='package.json')



from helper import CompatTestCase
from validator.compat import FX39_DEFINITION


class TestFX39Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 39 are properly executed."""

    VERSION = FX39_DEFINITION

    def test_noSuchMethod_literal(self):
        self.run_script_for_compat("""
            var o = {
              __noSuchMethod__: function(id, args) {
                console.log(id, '(' + args.join(', ') + ')');
              }
          };
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_noSuchMethod_assignment(self):
        self.run_script_for_compat("""
            var o = {};
            o.__noSuchMethod__ = function(id, args) {
              console.log(id, '(' + args.join(', ') + ')');
            };
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_noSuchMethod_string_assignment(self):
        self.run_script_for_compat("""
            var o = {};
            o['__noSuchMethod__'] = function(id, args) {
              console.log(id, '(' + args.join(', ') + ')');
            };
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_sendAsBinary(self):
        self.run_script_for_compat("""
           var xhr = new XMLHttpRequest();
           xhr.sendAsBinary('foo');
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_lightweightThemes_usedThemes(self):
        self.run_script_for_compat("""
           var value = getPermission('lightweightThemes.usedThemes');
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_lightweightThemes_isThemeSelected(self):
        self.run_script_for_compat("""
           var value = getPermission('lightweightThemes.isThemeSelected');
        """)
        self.assert_silent()
        self.assert_compat_error()



import re
import types

import actions
import jstypes
from validator.constants import BUGZILLA_BUG, EVENT_ASSIGNMENT


JS_URL = re.compile("href=[\'\"]javascript:")


def set_innerHTML(new_value, traverser):
    """Tests that values being assigned to innerHTML are not dangerous."""
    return _set_HTML_property('innerHTML', new_value, traverser)


def set_outerHTML(new_value, traverser):
    """Tests that values being assigned to outerHTML are not dangerous."""
    return _set_HTML_property('outerHTML', new_value, traverser)


# TODO(valcom): Make this generic and put it in utils
def _set_HTML_property(function, new_value, traverser):
    if not isinstance(new_value, jstypes.JSWrapper):
        new_value = jstypes.JSWrapper(new_value, traverser=traverser)

    if new_value.is_literal():
        literal_value = new_value.get_literal_value()
        if isinstance(literal_value, types.StringTypes):
            # Static string assignments

            HELP = ('Please avoid including JavaScript fragments in '
                    'HTML stored in JavaScript strings. Event listeners '
                    'should be added via `addEventListener` after the HTML '
                    'has been injected.',
                    'Injecting <script> nodes should be avoided when at all '
                    'possible. If you cannot avoid loading a script directly '
                    'into a content document, please consider doing so via '
                    'the subscript loader (http://mzl.la/1VGxOPC) instead. '
                    'If the subscript loader is not available, then the '
                    'script nodes should be created using `createElement`, '
                    'and should use a `src` attribute pointing to a '
                    '`resource:` URL within your extension.')

            # Test for on* attributes and script tags.
            if EVENT_ASSIGNMENT.search(literal_value.lower()):
                traverser.warning(
                    err_id=('testcases_javascript_instancetypes',
                            'set_%s' % function, 'event_assignment'),
                    warning='Event handler assignment via %s' % function,
                    description=('When assigning event handlers, %s '
                                 'should never be used. Rather, use a '
                                 'proper technique, like addEventListener.'
                                 % function,
                                 'Event handler code: %s'
                                 % literal_value.encode('ascii', 'replace')),
                    signing_help=HELP,
                    signing_severity='medium')
            elif ('<script' in literal_value or
                  JS_URL.search(literal_value)):
                traverser.warning(
                    err_id=('testcases_javascript_instancetypes',
                            'set_%s' % function, 'script_assignment'),
                    warning='Scripts should not be created with `%s`'
                            % function,
                    description='`%s` should not be used to add scripts to '
                                'pages via script tags or JavaScript URLs. '
                                'Instead, use event listeners and external '
                                'JavaScript.' % function,
                    signing_help=HELP,
                    signing_severity='medium')
            else:
                # Everything checks out, but we still want to pass it through
                # the markup validator. Turn off strict mode so we don't get
                # warnings about malformed HTML.
                from validator.testcases.markup.markuptester import (
                    MarkupParser)
                parser = MarkupParser(traverser.err, strict=False, debug=True)
                parser.process(traverser.filename, literal_value, 'xul')

    else:
        # Variable assignments
        traverser.warning(
            err_id=('testcases_javascript_instancetypes', 'set_%s' % function,
                    'variable_assignment'),
            warning='Markup should not be passed to `%s` dynamically.'
                    % function,
            description='Due to both security and performance concerns, '
                        '%s may not be set using dynamic values which have '
                        'not been adequately sanitized. This can lead to '
                        'security issues or fairly serious performance '
                        'degradation.' % function)


def set_on_event(new_value, traverser):
    """Ensure that on* properties are not assigned string values."""

    is_literal = new_value.is_literal()

    if (is_literal and
            isinstance(new_value.get_literal_value(), types.StringTypes)):
        traverser.warning(
            err_id=('testcases_javascript_instancetypes', 'set_on_event',
                    'on*_str_assignment'),
            warning='on* property being assigned string',
            description='Event handlers in JavaScript should not be '
                        'assigned by setting an on* property to a '
                        'string of JS code. Rather, consider using '
                        'addEventListener.',
            signing_help='Please add event listeners using the '
                         '`addEventListener` API. If the property you are '
                         'assigning to is not an event listener, please '
                         'consider renaming it, if at all possible.',
            signing_severity='medium')
    elif (not is_literal and isinstance(new_value.value, jstypes.JSObject) and
          'handleEvent' in new_value.value.data):
        traverser.warning(
            err_id=('js', 'on*', 'handleEvent'),
            warning='`handleEvent` no longer implemented in Gecko 18.',
            description='As of Gecko 18, objects with `handleEvent` methods '
                        'may no longer be assigned to `on*` properties. Doing '
                        'so will be equivalent to assigning `null` to the '
                        'property.')


def set__proto__(new_value, traverser):
    traverser.warning(
        err_id=('testcases_javascript_instanceproperties', '__proto__'),
        warning='Using __proto__ or setPrototypeOf to set a prototype is now '
                'deprecated.',
        description='Use of __proto__ or setPrototypeOf to set a prototype '
                    'causes severe performance degredation, and is '
                    'deprecated. You should use Object.create instead. '
                    'See bug %s for more information.' % BUGZILLA_BUG % 948227)


def set__exposedProps__(new_value, traverser):
    traverser.warning(
        err_id=('testcases_javascript_instanceproperties', '__exposedProps__'),
        warning='Use of deprecated __exposedProps__ declaration',
        description=(
            'The use of __exposedProps__ to expose objects to unprivileged '
            'scopes is dangerous, and has been deprecated. If objects '
            'must be exposed to unprivileged scopes, `cloneInto` or '
            '`exportFunction` should be used instead.'),
        signing_help='If you are using this API to expose APIs to content, '
                     'please use `Components.utils.cloneInto`, or '
                     '`Components.utils.exportFunction` '
                     '(http://mzl.la/1fvvgm9). If you are using it '
                     'for other purposes, please consider using a built-in '
                     'message passing interface instead. Extensions which '
                     'expose APIs to content will be required to go through '
                     'manual code review for at least one submission.',
        signing_severity='high')


def set_contentScript(value, traverser):
    """Warns when values are assigned to the `contentScript` properties,
    which are essentially the same as calling `eval`."""

    if value.is_literal():
        content_script = actions._get_as_str(value)

        # Avoid import loop.
        from validator.testcases.scripting import test_js_file
        test_js_file(
            traverser.err, traverser.filename, content_script,
            line=traverser.line, context=traverser.context)
    else:
        traverser.warning(
            err_id=('testcases_javascript_instanceproperties',
                    'contentScript', 'set_non_literal'),
            warning='`contentScript` properties should not be used',
            description='Creating content scripts from dynamic values '
                        'is dangerous and error-prone. Please use a separate '
                        'JavaScript file, along with the '
                        '`contentScriptFile` property instead.',
            signing_help='Please do not use the `contentScript` property '
                         'in any add-ons submitted for automated signing.',
            signing_severity='high')


OBJECT_DEFINITIONS = {
    'innerHTML': {'set': set_innerHTML},
    'outerHTML': {'set': set_outerHTML},
    'contentScript': {'set': set_contentScript},
    '__proto__': {'set': set__proto__},
    '__exposedProps__': {'set': set__exposedProps__},
}


def get_operation(mode, prop):
    """
    This returns the object definition function for a particular property
    or mode. mode should either be 'set' or 'get'.
    """

    if prop in OBJECT_DEFINITIONS and mode in OBJECT_DEFINITIONS[prop]:
        return OBJECT_DEFINITIONS[prop][mode]

    elif mode == 'set' and unicode(prop).startswith('on'):
        # We can't match all of them manually, so grab all the "on*" properties
        # and funnel them through the set_on_event function.

        return set_on_event



from helper import CompatTestCase
from validator.compat import FX42_DEFINITION


class TestFX42Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 42 are properly executed."""

    VERSION = FX42_DEFINITION

    def test_parseContentType(self):
        self.run_script_for_compat("""
            var netutil = Components
              .classes["@mozilla.org/network/util;1"]
              .getService(Components.interfaces.nsINetUtil);
            netutil.parseContentType('text/html', 'utf-8', {});
        """)
        self.assert_compat_error()

    def test_mozRequestAnimationFrame(self):
        self.run_script_for_compat("""
            // Don't flag comments: window.mozRequestAnimationFrame()
            window.requestAnimationFrame();
        """)
        assert not self.compat_err.errors
        assert not self.compat_err.warnings

        self.run_script_for_compat("""
            window.mozRequestAnimationFrame();
        """)
        self.assert_compat_error()

    def _test_nsIPermissionManager(self, method):
        self.run_script_for_compat("""
            var perms = Cc["@mozilla.org/permissionmanager;1"]
                          .getService(Ci.nsIPermissionManager);
            %s
        """ % method)
        self.assert_compat_error()

    def _test_ServicesPerms(self, method):
        self.run_script_for_compat("""
            Components.utils.import("resource://gre/modules/Services.jsm");
            var perms = Services.perms;
            %s
        """ % method)
        self.assert_compat_error()

    def test_nsIPermissionManagerMethods(self):
        methods = ("perms.add(url, 'cookie', 1);",
                   "perms.addFromPrincipal(url, 'cookie', 1);",
                   "perms.remove('yahoo.com', 'cookie');",
                   "perms.removeFromPrincipal('yahoo.com', 'cookie');",
                   "perms.removeAll();",
                   "perms.testExactPermission(url, cookie);"
                   "perms.testExactPermissionFromPrincipal(url, cookie);",
                   "perms.testPermission(url, cookie);",
                   "perms.testPermissionFromPrincipal(url, cookie);")

        for method in methods:
            yield self._test_nsIPermissionManager, method

        for method in methods:
            yield self._test_ServicesPerms, method



from collections import defaultdict
import hashlib
from itertools import imap
import json
import os
import pickle

import validator.decorator as decorator
from validator.constants import PACKAGE_EXTENSION
from validator.version import Version
from content import FLAGGED_FILES


SAFE_FILES = ('.jpg', '.ico', '.png', '.gif', '.txt')

JETPACK_DATA_FILE = os.path.join(os.path.dirname(__file__), 'jetpack_data.txt')
JETPACK_PICKLE_FILE = JETPACK_DATA_FILE + '.pickle'

if os.path.exists(JETPACK_PICKLE_FILE):
    with open(JETPACK_PICKLE_FILE) as pickle_file:
        jetpack_hash_table = pickle.load(pickle_file)
        latest_jetpack = pickle.load(pickle_file)
else:
    # Read the jetpack data file in.
    with open(JETPACK_DATA_FILE) as jetpack_data:
        # Parse the jetpack data into something useful.
        jetpack_hash_table = defaultdict(dict)
        latest_jetpack = None
        for path, version_str, hash in imap(str.split, jetpack_data):
            version = Version(version_str)
            if version.is_release and (not latest_jetpack or
                                       version > latest_jetpack):
                latest_jetpack = version
            jetpack_hash_table[hash][version_str] = path

    with open(JETPACK_PICKLE_FILE, 'w') as pickle_file:
        pickle.dump(jetpack_hash_table, pickle_file)
        pickle.dump(latest_jetpack, pickle_file)


@decorator.register_test(tier=1, expected_type=PACKAGE_EXTENSION)
def inspect_jetpack(err, xpi_package, allow_old_sdk=False):
    """
    If the add-on is a Jetpack extension, its contents should be tested to
    ensure that none of the Jetpack libraries have been tampered with.
    """

    if not is_jetpack(xpi_package):
        return

    err.metadata['is_jetpack'] = True
    if not err.get_resource('pretested_files'):
        err.save_resource('pretested_files', [])

    if is_old_jetpack(xpi_package):
        # Ignore this test for cfx based add-ons if run as part of bulk
        # validation because existing add-ons on AMO have been re-packed.
        # We just need this for validation during upload.
        if err.get_resource('is_compat_test'):
            return
        err.error(
            err_id=('jetpack', 'inspect_jetpack',
                    'cfx'),
            error='Add-ons built with "cfx" are no longer accepted.',
            description='Due to breaking changes in Firefox 44, add-ons packed '
                        'with the "cfx" tool won\'t work any longer. To make '
                        'your add-on compatible again, please use the "jpm" '
                        'tool (http://mzl.la/1PwF5fY). You can find steps to '
                        'migrate from "cfx" to "jpm" at http://mzl.la/1ncmH3A')
        return
    else:
        metadata_validator = PackageJsonValidator(err, xpi_package)

    # Prepare a place to store mentioned hashes.
    found_hashes = set()
    unknown_files = []
    tested_files = metadata_validator.tested_files
    file_hashes = metadata_validator.file_hashes
    pretested_files = err.get_resource('pretested_files')

    # Iterate the rest of the files in the package for testing.
    for filename in xpi_package:

        # Skip files we've already identified.
        if (filename in tested_files or
                filename == 'harness-options.json'):
            continue

        # Skip safe files.
        if (filename.lower().endswith(SAFE_FILES) or
                filename in FLAGGED_FILES):
            continue

        file_data = xpi_package.read(filename)
        # Skip empty files.
        if not file_data.strip():
            continue
        blob_hash = (file_hashes.get(filename, None) or
                     hashlib.sha256(file_data).hexdigest())
        file_hashes[filename] = blob_hash

        if blob_hash not in jetpack_hash_table:
            unknown_files.append(filename)
            continue
        else:
            tested_files[filename] = jetpack_hash_table[blob_hash]
            pretested_files.append(filename)

        # Mark the hashes we actually find as being present.
        if blob_hash in found_hashes:
            found_hashes.discard(blob_hash)

    # Store the collected information in the output metadata.
    err.save_resource('pretested_files', pretested_files)
    err.metadata['jetpack_unknown_files'] = unknown_files

    if metadata_validator.suppress_warnings:
        return

    sdk_version = err.metadata.get('jetpack_sdk_version')
    for zip_path, versions in tested_files.items():
        if sdk_version in versions:
            tested_files[zip_path] = sdk_version, versions[sdk_version]
        else:
            # This isn't the version it claims to be. Go with the latest
            # Jetpack version we know about that contains it.
            version = str(max(Version(v) for v in versions))
            tested_files[zip_path] = version, versions[version]

            if (not zip_path.endswith('/') and
                    not metadata_validator.use_latest_version):
                err.warning(
                    err_id=('testcases_jetpack',
                            'inspect_jetpack',
                            'mismatched_version'),
                    warning='Jetpack module version mismatch',
                    description=('A file in the Jetpack add-on does not match '
                                 'the SDK version specified in harness-options'
                                 '.json.',
                                 'Module: %s' % zip_path,
                                 'Versions: %s/%s' % (sdk_version, version)),
                    filename=zip_path)

    # We've got hashes left over
    if found_hashes:
        err.warning(
            err_id=('testcases_jetpack', 'inspect_jetpack', 'extra_hashes'),
            warning='Extra hashes registered in harness-options.',
            description=('This Jetpack add-on registers modules in the '
                         'harness-options.json file that do not exist in the '
                         'package.',
                         'Hashes: %s' % ', '.join(found_hashes)),
            filename='harness-options.json')

    # Store the collected information in the output metadata.
    err.metadata['jetpack_identified_files'] = tested_files

    identified_files = err.metadata.setdefault('identified_files', {})

    for file, (version, path) in tested_files.items():
        identified_files[file] = {'path': path, 'version': version,
                                  'library': 'Jetpack'}


def is_jetpack(xpi):
    return has_bootstrap(xpi) and (has_harness_options(xpi) or
                                   has_package_json(xpi))


def is_old_jetpack(xpi):
    return has_bootstrap(xpi) and has_harness_options(xpi)


def has_bootstrap(xpi):
    return 'bootstrap.js' in xpi


def has_harness_options(xpi):
    return 'harness-options.json' in xpi


def has_package_json(xpi):
    return 'package.json' in xpi


class PackageJsonValidator(object):
    def __init__(self, err, xpi_package):
        self.tested_files = {}
        self.file_hashes = {}
        self.suppress_warnings = False
        self.use_latest_version = True



from helper import CompatTestCase
from validator.compat import FX43_DEFINITION


class TestFX43Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 38 are properly executed."""

    VERSION = FX43_DEFINITION

    def test_mozFetchAsStream(self):
        self.run_script_for_compat('canvas.mozFetchAsStream(aCallback, this.contentType);')
        self.assert_silent()
        self.assert_compat_error()



import json
import os
import types

from . import constants
from .constants import PACKAGE_ANY
from .errorbundler import ErrorBundle
# This is necessary. Do not remove it unless you know exactly what
# you are doing.
import loader  # noqa
import submain


def validate(path, format='json',
             approved_applications=None,
             determined=True,
             listed=True,
             expectation=PACKAGE_ANY,
             for_appversions=None,
             overrides=None,
             timeout=-1,
             compat_test=False,
             **kw):
    """
    Perform validation in one easy step!

    `path`:
        *Required*
        A file system path to the package to be validated.
    `format`:
        The format to return the results in. Defaults to "json". Currently, any
        other format will simply return the error bundle.
    `approved_applications`:
        Path to the list of approved application versions
    `determined`:
        If set to `False`, validation will halt at the end of the first tier
        that raises errors.
    `listed`:
        Whether the app is headed for the app marketplace or AMO. Defaults to
        `True`.
    `expectation`:
        The type of package that should be expected. Must be a symbolic
        constant from validator.constants (i.e.:
        validator.constants.PACKAGE_*). Defaults to PACKAGE_ANY.
    `for_appversions`:
        A dict of app GUIDs referencing lists of versions. Determines which
        version-dependant tests should be run.
    `timeout`:
        Number of seconds before aborting addon validation, or -1 to
        run with no timeout.
    `compat_tests`:
        A flag to signal the validator to skip tests which should not be run
        during compatibility bumps. Defaults to `False`.
    """

    bundle = ErrorBundle(listed=listed, determined=determined,
                         overrides=overrides, for_appversions=for_appversions)
    bundle.save_resource('is_compat_test', compat_test)

    if approved_applications is None:
        approved_applications = os.path.join(os.path.dirname(__file__),
                                             'app_versions.json')

    if isinstance(approved_applications, types.StringTypes):
        # Load up the target applications if the approved applications is a
        # path (string).
        with open(approved_applications) as approved_apps:
            apps = json.load(approved_apps)
    elif isinstance(approved_applications, dict):
        # If the lists of approved applications are already in a dict, just use
        # that instead of trying to pull from a file.
        apps = approved_applications
    else:
        raise ValueError('Unknown format for `approved_applications`.')

    constants.APPROVED_APPLICATIONS.clear()
    constants.APPROVED_APPLICATIONS.update(apps)

    submain.prepare_package(bundle, path, expectation,
                            for_appversions=for_appversions,
                            timeout=timeout)

    return format_result(bundle, format)


def format_result(bundle, format):
    # Write the results to the pipe
    formats = {'json': lambda b: b.render_json()}
    if format is not None:
        return formats[format](bundle)
    else:
        return bundle



from helper import CompatTestCase
from validator.compat import FX41_DEFINITION


class TestFX41Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 41 are properly executed."""

    VERSION = FX41_DEFINITION

    def test_browser_newtab_url_pref(self):
        self.run_script_for_compat("""
            require("sdk/preferences/service").set("browser.newtab.url", false);
        """)
        self.assert_failed(with_warnings=[{
            "message": "Potentially unsafe preference branch referenced",
            "signing_severity": "high",
        }])
        self.assert_compat_error()



from functools import partial
import math

import actions
from actions import _get_as_str
import call_definitions
from call_definitions import xpcom_constructor as xpcom_const, python_wrap
from entity_values import entity
import instanceactions
from jstypes import JSWrapper
from validator.compat import FX40_DEFINITION
from validator.constants import MDN_DOC


# A list of identifiers and member values that may not be used.
BANNED_IDENTIFIERS = {
    u'newThread':
        'Creating threads from JavaScript is a common cause '
        'of crashes and is unsupported in recent versions of the platform',
    u'processNextEvent':
        'Spinning the event loop with processNextEvent is a common cause of '
        'deadlocks, crashes, and other errors due to unintended reentrancy. '
        'Please use asynchronous callbacks instead wherever possible',
}

CUSTOMIZATION_API_HELP = (
    'We are currently working to provide libraries and APIs to allow '
    'extensions to modify these settings in ways that we can guarantee are '
    'in-policy. In the interim, we recommend that you avoid changing these '
    'settings altogether, if at all possible.')

CUSTOMIZATION_PREF_MESSAGE = {
    'description': (
        'Extensions must not alter user preferences such as the current home '
        'page, new tab page, or search engine, without explicit user consent, '
        'in which a user takes a non-default action. Such changes must also '
        'be reverted when the extension is disabled or uninstalled.',
        'In nearly all cases, new values for these preferences should be '
        'set in the default preference branch, rather than the user branch.'),
    'signing_help':
        'Add-ons which directly change these preferences must undergo at '
        'manual code review for at least one submission. ' +
        CUSTOMIZATION_API_HELP,
    'signing_severity': 'high',
}

NETWORK_PREF_MESSAGE = {
    'description':
        'Changing network preferences may be dangerous, and often leads to '
        'performance costs.',
    'signing_help':
        'Changes to these preferences are strongly discouraged. If at all '
        'possible, you should remove any reference to them from '
        'your extension. Extensions which do modify these preferences '
        'must undergo light manual code review for at least one submission.',
    'signing_severity': 'low',
}

SEARCH_PREF_MESSAGE = {
    'description':
        'Search engine preferences may not be changed by add-ons directly. '
        'All such changes must be made only via the browser search service, '
        'and only after an explicit opt-in from the user. All such changes '
        'must be reverted when the extension is disabled or uninstalled.',
    'signing_help': (
        'You should remove all references to these preferences from your '
        'code, and interact with search settings only via the '
        '`Services.search` interface. Extensions which interact with these '
        'preferences directly are not acceptable within the Firefox add-on '
        'ecosystem.',
        'Note, however, that extensions which change search settings even via '
        'the search service must undergo manual code review for at least '
        'one submission. ' + CUSTOMIZATION_API_HELP),
    'signing_severity': 'high',
}

SECURITY_PREF_MESSAGE = {
    'description':
        'Changing this preference may have severe security implications, and '
        'is forbidden under most circumstances.',
    'editors_only': True,
    'signing_help': ('Extensions which alter these settings are allowed '
                     'within the Firefox add-on ecosystem by exception '
                     'only, and under extremely limited circumstances.',
                     'Please remove any reference to these preference names '
                     'from your add-on.'),
    'signing_severity': 'high',
}

MARIONETTE_MESSAGE = {
    'warning': 'Marionette should not be accessed by extensions',
    'description': 'References to the Marionette service are not acceptable '
                   'in extensions. Please remove them.',
}


def fuel_error(traverse_node, err):
    traverse_node.im_self.warning(
        err_id=('js', 'traverser', 'dangerous_global'),
        warning='The FUEL library is now deprecated.',
        description='The FUEL library is now deprecated. You should use the '
                    'add-ons SDK or Services.jsm. See %s for more information.'
                    % MDN_DOC % 'Mozilla/Tech/Toolkit_API/FUEL',
        for_appversions=FX40_DEFINITION,
        tier=5,
        compatibility_type='warning')


BANNED_PREF_BRANCHES = (
    # Security and update preferences
    (u'app.update.', SECURITY_PREF_MESSAGE),
    (u'browser.addon-watch.', SECURITY_PREF_MESSAGE),
    (u'capability.policy.', None),
    (u'datareporting.', SECURITY_PREF_MESSAGE),

    (u'extensions.blocklist.', SECURITY_PREF_MESSAGE),
    (u'extensions.checkCompatibility', None),
    (u'extensions.getAddons.', SECURITY_PREF_MESSAGE),
    (u'extensions.update.', SECURITY_PREF_MESSAGE),
    (u'xpinstall.signatures.required', SECURITY_PREF_MESSAGE),

    # Let's see if we can get away with this...
    # Changing any preference in this branch should result in a
    # warning. However, this substring may turn out to be too
    # generic, and lead to spurious warnings, in which case we'll
    # have to single out sub-branches.
    (u'security.', SECURITY_PREF_MESSAGE),

    # Search, homepage, and customization preferences
    (u'browser.newtab.url', CUSTOMIZATION_PREF_MESSAGE),
    (u'browser.newtabpage.enabled', CUSTOMIZATION_PREF_MESSAGE),
    (u'browser.search.defaultenginename', SEARCH_PREF_MESSAGE),
    (u'browser.search.searchEnginesURL', SEARCH_PREF_MESSAGE),
    (u'browser.startup.homepage', CUSTOMIZATION_PREF_MESSAGE),
    (u'extensions.getMoreThemesURL', None),
    (u'keyword.URL', SEARCH_PREF_MESSAGE),
    (u'keyword.enabled', SEARCH_PREF_MESSAGE),

    # Network
    (u'network.proxy.autoconfig_url', {
        'description':
            'As many add-ons have reason to change the proxy autoconfig URL, '
            'and only one at a time may do so without conflict, extensions '
            'must make proxy changes using other mechanisms. Installing a '
            'proxy filter is the recommended alternative: '
            'https://developer.mozilla.org/en-US/docs/Mozilla/Tech/XPCOM/'
            'Reference/Interface/nsIProtocolProxyService#registerFilter()',
        'signing_help':
            'Dynamic proxy configuration should be implemented via proxy '
            'filters, as described above. This preference should not be '
            'set, except directly by end users.',
        'signing_severity': 'low'}),
    (u'network.proxy.', NETWORK_PREF_MESSAGE),
    (u'network.http.', NETWORK_PREF_MESSAGE),
    (u'network.websocket.', NETWORK_PREF_MESSAGE),

    # Other
    (u'browser.preferences.instantApply', None),

    (u'extensions.alwaysUnpack', None),
    (u'extensions.bootstrappedAddons', None),
    (u'extensions.dss.', None),
    (u'extensions.installCache', None),
    (u'extensions.lastAppVersion', None),
    (u'extensions.pendingOperations', None),

    (u'general.useragent.', None),

    (u'nglayout.debug.disable_xul_cache', None),

    # Marionette
    (u'marionette.force-local', MARIONETTE_MESSAGE),
    (u'marionette.defaultPrefs.enabled', MARIONETTE_MESSAGE),
    (u'marionette.defaultPrefs.port', MARIONETTE_MESSAGE),
)

BANNED_PREF_REGEXPS = [
    r'extensions\..*\.update\.(url|enabled|interval)',
]


def is_shared_scope(traverser, right=None, node_right=None):
    """Returns true if the traverser `t` is traversing code loaded into
    a shared scope, such as a browser window. Particularly used for
    detecting when global overwrite warnings should be issued."""

    # FIXME(Kris): This is not a great heuristic.
    return not (traverser.is_jsm or
                traverser.err.get_resource('em:bootstrap') == 'true')


# See https://github.com/mattbasta/amo-validator/wiki/JS-Predefined-Entities
# for details on entity properties.

CONTENT_DOCUMENT = None


CATEGORY_MANAGER = {
    u'addCategoryEntry':
        {'dangerous':
            lambda a, t, e:
                e.get_resource('em:bootstrap') and
                ('Bootstrapped add-ons may not create persistent category '
                 'entries.' if len(a) > 3 and t(a[3]).is_literal() else
                 'Authors of bootstrapped add-ons must take care to clean up '
                 'any added category entries at shutdown.')}}


OBSOLETE_EXTENSION_MANAGER = {
    'value': {},
    'dangerous': 'This interface is part of the obsolete extension manager '
                 'interface, which is not available in any remotely modern '
                 'version of Firefox. It should not be referenced in any '
                 'code.'}


ADDON_INSTALL_METHOD = {
    'value': {},
    'dangerous': {
        'description': (
            'Add-ons may install other add-ons only by user consent. Any '
            'such installations must be carefully reviewed to ensure '
            'their safety.'),
        'editors_only': True,
        'signing_help': (
            "Rather than directly install other add-ons, you should offer "
            "users the opportunity to install them via the normal web install "
            "process, using an install link or button connected to the "
            "`InstallTrigger` API: "
            "https://developer.mozilla.org/en-US/docs/Web/API/InstallTrigger",
            "Updates to existing add-ons should be provided via the "
            "manifest's `updateURL` mechanism."),
        'signing_severity': 'high'},
}


SEARCH_MESSAGE = 'Potentially dangerous use of the search service'
SEARCH_DESCRIPTION = (
    'Changes to the default and currently-selected search engine settings '
    'may only take place after users have explicitly opted-in, by taking '
    'a non-default action. Any such changes must be reverted when the add-on '
    'making them is disabled or uninstalled.')

def search_warning(severity='medium', editors_only=False,
                   message=SEARCH_MESSAGE,
                   description=SEARCH_DESCRIPTION):
    return {'err_id': ('testcases_javascript_actions',
                       'search_service',
                       'changes'),
            'signing_help':
                'Add-ons which directly change search settings must undergo '
                'manual code review for at least one submission. ' +
                CUSTOMIZATION_API_HELP,
            'signing_severity': severity,
            'editors_only': editors_only,
            'warning': message,
            'description': description}


REGISTRY_WRITE = {'dangerous': {
    'err_id': ('testcases_javascript_actions',
               'windows_registry',
               'write'),
    'warning': 'Writes to the registry may be dangerous',
    'description': 'Writing to the registry can have many system-level '
                   'consequences and requires careful review.',
    'signing_help': (
        'Please store any settings relevant to your add-on within the '
        'current Firefox profile, ideally using the preferences service.'
        'If you are intentionally changing system settings, consider '
        'searching for a Firefox API which has a similar effect. If no such '
        'API exists, we strongly discourage making any changes which affect '
        'the system outside of the browser.'),
    'signing_severity': 'medium',
    'editors_only': True}}


def registry_key(write=False):
    """Represents a function which returns a registry key object."""
    res = {'return': lambda wrapper, arguments, traverser: (
        build_quick_xpcom('createInstance', 'nsIWindowMediator',
                          traverser, wrapper=True))}
    if write:
        res.update(REGISTRY_WRITE)

    return res


INTERFACES = {
    u'nsISupports': {'value': {}},
    u'mozIStorageBaseStatement':
        {'value':
            {u'execute':
                {'dangerous': instanceactions.SYNCHRONOUS_SQL_DESCRIPTION},
             u'executeStep':
                {'dangerous': instanceactions.SYNCHRONOUS_SQL_DESCRIPTION}}},
    u'nsIExtensionManager': OBSOLETE_EXTENSION_MANAGER,
    u'nsIUpdateItem': OBSOLETE_EXTENSION_MANAGER,
    u'nsIInstallLocation': OBSOLETE_EXTENSION_MANAGER,
    u'nsIAddonInstallListener': OBSOLETE_EXTENSION_MANAGER,
    u'nsIAddonUpdateCheckListener': OBSOLETE_EXTENSION_MANAGER,
    u'nsICategoryManager':
        {'value': CATEGORY_MANAGER},
    u'nsIAccessibleRetrieval':
        {'dangerous':
            'Using the nsIAccessibleRetrieval interface causes significant '
            'performance degradation in Gecko. It should only be used in '
            'accessibility-related add-ons.',
         'value': {}},
    u'nsIBrowserSearchService':
        {'value':
            {u'currentEngine':
                {'readonly': search_warning(severity='high')},
             u'defaultEngine':
                {'readonly': search_warning(severity='high')},
             u'addEngine':
                {'dangerous': search_warning()},
             u'addEngineWithDetails':
                {'dangerous': search_warning()},
             u'removeEngine':
                {'dangerous': search_warning()},
             u'moveEngine':
                {'dangerous': search_warning()}}},

    u'nsIComponentRegistrar':
        {'value':
            {u'autoRegister':
                {'dangerous':
                    lambda a, t, e:
                        e.get_resource('em:bootstrap') and
                        'Bootstrapped add-ons may not register chrome '
                        'manifest files.'},
             u'registerFactory':
                {'dangerous':
                    lambda a, t, e:
                        e.get_resource('em:bootstrap') and
                        'Authors of bootstrapped add-ons must take care to '
                        'clean up any component registrations at shutdown.'}}},
    u'nsIDNSService': {'value': {u'resolve': entity('nsIDNSService.resolve')}},
    u'nsIJSON':
        {'value':
            {u'encode':
                {'return': call_definitions.nsIJSON_deprec},
             u'decode':
                {'return': call_definitions.nsIJSON_deprec}}},
    u'nsIMsgDatabase':
        {'value':
            {u'forceFolderDBClosed': entity('nsIMsgDatabase.forceFolderDBClosed')}},
    u'nsIObserverService':
        {'value':
            {u'addObserver': entity('nsIObserverService.addObserver')},
         'dangerous': lambda a, t, e:
            lambda t, e: (
                e.metadata.get('is_jetpack') and
                'The observer service should not be used directly in SDK '
                "add-ons. Please use the 'sdk/system/events' module "
                'instead.')},
    u'nsIPrefBranch':
        {'value': dict(
            tuple((method, {'return': instanceactions.set_preference})
                   for method in (u'setBoolPref',
                                  u'setCharPref',
                                  u'setComplexValue',
                                  u'setIntPref',
                                  u'clearUserPref',
                                  u'deleteBranch',
                                  u'resetBranch')) +
            tuple((method, {'return': instanceactions.get_preference})
                   for method in (u'getBoolPref',
                                  u'getCharPref',
                                  u'getChildList',
                                  u'getComplexValue',
                                  u'getFloatPref',
                                  u'getIntPref',
                                  u'getPrefType',
                                  u'prefHasUserValue')))},
    u'nsIResProtocolHandler':
        {'value':
            {u'setSubstitution':
                {'dangerous':
                    lambda a, t, e:
                        e.get_resource('em:bootstrap') and
                        a and len(a) > 1 and t(a[1]).get_literal_value() and
                        'Authors of bootstrapped add-ons must take care '
                        'to clean up any added resource substitutions '
                        'at shutdown.'}}},
    u'nsISound': {'value': {'play': entity('nsISound.play')}},
    u'nsIStringBundleService':
        {'value':
            {u'createStringBundle':
                {'dangerous':
                    lambda a, t, e:
                        e.get_resource('em:bootstrap') and
                        'Authors of bootstrapped add-ons must take care '
                        'to flush the string bundle cache at shutdown.'},
             u'createExtensibleBundle':
                {'dangerous':
                    lambda a, t, e:
                        e.get_resource('em:bootstrap') and
                        'Authors of bootstrapped add-ons must take care '
                        'to flush the string bundle cache at shutdown.'}}},
    u'nsIStyleSheetService':
        {'value':
            {u'loadAndRegisterSheet':
                {'dangerous':
                    lambda a, t, e:
                        e.get_resource('em:bootstrap') and
                        'Authors of bootstrapped add-ons must take care to '
                        'unregister registered stylesheets at shutdown.'}}},
    u'nsITransferable':
        {'value':
            {u'init':
                entity('nsITransferable.init')}},
    u'nsIWindowMediator':
        {'value':
            {'registerNotification':
                {'dangerous':
                    lambda a, t, e:
                        e.get_resource('em:bootstrap') and
                        'Authors of bootstrapped add-ons must take care '
                        'to remove any added observers at shutdown.'}}},
    u'nsIWindowWatcher':
        {'value':
            {u'addListener':
                {'dangerous':
                    lambda a, t, e:
                        e.get_resource('em:bootstrap') and
                        'Authors of bootstrapped add-ons must take care '
                        'to remove any added observers at shutdown.'},
             u'openWindow': entity('nsIWindowWatcher.openWindow')}},
    u'nsIProtocolProxyService': {'value': {
        u'registerFilter': {'dangerous': {
            'err_id': ('testcases_javascript_actions',
                       'predefinedentities', 'proxy_filter'),
            'description': (
                'Proxy filters can be used to direct arbitrary network '
                'traffic through remote servers, and may potentially '
                'be abused.',
                'Additionally, to prevent conflicts, the `applyFilter` '
                'method should always return its third argument in cases '
                'when it is not supplying a specific proxy.'),
            'editors_only': True,
            'signing_help': 'Due to the potential for unintended effects, '
                            'any add-on which uses this API must undergo '
                            'manual code review for at least one submission.',
            'signing_severity': 'low'}}}},
    u'nsIWebBrowserPersist':
        {'value':
             {u'saveChannel':
                  {'return': call_definitions.webbrowserpersist},
              u'saveURI':
                  {'return':
                       call_definitions.webbrowserpersist_saveuri},
              u'savePrivacyAwareURI':
                  {'return': call_definitions.webbrowserpersist}}},
    u'nsIMsgCompose':
        {'value':
             {u'checkAndPopulateRecipients': entity('nsIMsgCompose.checkAndPopulateRecipients')}},
    u'nsIFolderLookupService':
        {'value':
             {u'getFolderById': entity('nsIFolderLookupService.getFolderById')}},
    u'nsIAbCard':
        {'value':
             {u'kAllowRemoteContentProperty': entity('nsIAbCard.kAllowRemoteContentProperty')}},
    u'nsIAddrDatabase':
        {'value':
             {u'addAllowRemoteContent': entity('nsIAddrDatabase.addAllowRemoteContent')}},

    'nsIWindowsRegKey': {'value': {u'create': REGISTRY_WRITE,
                                   u'createChild': registry_key(write=True),
                                   u'openChild': registry_key(),
                                   u'writeBinaryValue': REGISTRY_WRITE,
                                   u'writeInt64Value': REGISTRY_WRITE,
                                   u'writeIntValue': REGISTRY_WRITE,
                                   u'writeStringValue': REGISTRY_WRITE,
                                  }},
    'nsIPermissionManager': {'value': {
                                'add': entity('nsIPermissionManager.add'),
                                'addFromPrincipal': entity('nsIPermissionManager.addFromPrincipal'),
                                'remove': entity('nsIPermissionManager.remove'),
                                'removeFromPrincipal': entity('nsIPermissionManager.removeFromPrincipal'),
                                'removeAll': entity('nsIPermissionManager.removeAll'),
                                'testExactPermission': entity('nsIPermissionManager.testExactPermission'),
                                'testExactPermissionFromPrincipal': entity('nsIPermissionManager.testExactPermissionFromPrincipal'),
                                'testPermission': entity('nsIPermissionManager.testPermission'),
                                'testPermissionFromPrincipal': entity('nsIPermissionManager.testPermissionFromPrincipal')}},
}

INTERFACE_ENTITIES = {u'nsIXMLHttpRequest':
                          {'xpcom_map':
                               lambda: GLOBAL_ENTITIES['XMLHttpRequest']},
                      u'nsIProcess': {'dangerous': {
                        'warning': 'The use of nsIProcess is potentially '
                                   'dangerous and requires careful review '
                                   'by an administrative reviewer.',
                        'editors_only': True,
                        'signing_help':
                          'Consider alternatives to directly launching '
                          'executables, such as loading a URL with an '
                          'appropriate external protocol handler, making '
                          'network requests to a local service, or using '
                          'the (as a last resort) `nsIFile.launch()` method '
                          'to open a file with the appropriate application.',
                        'signing_severity': 'high',
                      }},
                      u'nsIDOMGeoGeolocation': {'dangerous':
                         'Use of the geolocation API by add-ons requires '
                         'prompting users for consent.'},
                      u'nsIWindowsRegKey': {'dangerous': {
                          'signing_help':
                            'The information stored in many standard registry '
                            'keys is available via built-in Firefox APIs, '
                            'such as `Services.sysinfo`, `Services.dirsvc`, '
                            'and the environment service '
                            '(http://mzl.la/1OGgCF3). We strongly discourage '
                            'extensions from reading registry information '
                            'which is not available via other Firefox APIs.',
                          'signing_severity': 'low',
                          'editors_only': True,
                          'description': (
                              'Access to the registry is potentially '
                              'dangerous, and should be reviewed with special '
                              'care.')}},
                      }

DANGEROUS_CERT_DB = {
    'err_id': ('javascript', 'predefinedentities', 'cert_db'),
    'description': 'Access to the X509 certificate '
                   'database is potentially dangerous '
                   'and requires careful review by an '
                   'administrative reviewer.',
    'editors_only': True,
    'signing_help': 'Please avoid interacting with the certificate and trust '
                    'databases if at all possible. Any add-ons which interact '
                    'with these databases must undergo manual code review '
                    'prior to signing.',
    'signing_severity': 'high',
}

INTERFACE_ENTITIES.update(
    (interface, {'dangerous': DANGEROUS_CERT_DB})
    for interface in ('nsIX509CertDB', 'nsIX509CertDB2', 'nsIX509CertList',
                      'nsICertOverrideService'))

CONTRACT_ENTITIES = {
    contract: DANGEROUS_CERT_DB
    for contract in ('@mozilla.org/security/x509certdb;1',
                     '@mozilla.org/security/x509certlist;1',
                     '@mozilla.org/security/certoverride;1')}

for interface in INTERFACES:
    def construct(interface):
        def wrap():
            return INTERFACES[interface]
        return wrap
    if interface not in INTERFACE_ENTITIES:
        INTERFACE_ENTITIES[interface] = {}
    INTERFACE_ENTITIES[interface]['xpcom_map'] = construct(interface)


def build_quick_xpcom(method, interface, traverser, wrapper=False):
    """A shortcut to quickly build XPCOM objects on the fly."""
    extra = ()
    if isinstance(interface, (list, tuple)):
        interface, extra = interface[0], interface[1:]

    def interface_obj(iface):
        return traverser._build_global(
            name=method,
            entity={'xpcom_map':
                lambda: INTERFACES.get(iface, INTERFACES['nsISupports'])})

    constructor = xpcom_const(method, pretraversed=True)
    obj = constructor(None, [interface_obj(interface)], traverser)

    for iface in extra:
        # `xpcom_constructor` really needs to be cleaned up so we can avoid
        # this duplication.
        iface = interface_obj(iface)
        iface = traverser._build_global('QueryInterface',
                                        iface.value['xpcom_map']())

        obj.value = obj.value.copy()

        value = obj.value['value'].copy()
        value.update(iface.value['value'])

        obj.value.update(iface.value)
        obj.value['value'] = value

    if isinstance(obj, JSWrapper) and not wrapper:
        obj = obj.value
    return obj


UNSAFE_TEMPLATE_METHOD = (
    'The use of `%s` can lead to unsafe '
    'remote code execution, and therefore must be done with '
    'great care, and only with sanitized data.')


SERVICES = {
    'appinfo': ('nsIXULAppInfo', 'nsIXULRuntime'),
    'appShell': 'nsIAppShellService',
    'blocklist': 'nsIBlocklistService',
    'cache': 'nsICacheService',
    'cache2': 'nsICacheStorageService',
    'clipboard': 'nsIClipboard',
    'console': 'nsIConsoleService',
    'contentPrefs': 'nsIContentPrefService',
    'cookies': ('nsICookieManager', 'nsICookieManager2', 'nsICookieService'),
    'dirsvc': ('nsIDirectoryService', 'nsIProperties'),
    'DOMRequest': 'nsIDOMRequestService',
    'domStorageManager': 'nsIDOMStorageManager',
    'downloads': 'nsIDownloadManager',
    'droppedLinkHandler': 'nsIDroppedLinkHandler',
    'eTLD': 'nsIEffectiveTLDService',
    'focus': 'nsIFocusManager',
    'io': ('nsIIOService', 'nsIIOService2'),
    'locale': 'nsILocaleService',
    'logins': 'nsILoginManager',
    'obs': 'nsIObserverService',
    'perms': 'nsIPermissionManager',
    'prefs': ('nsIPrefBranch2', 'nsIPrefService', 'nsIPrefBranch'),
    'prompt': 'nsIPromptService',
    'scriptloader': 'mozIJSSubScriptLoader',
    'scriptSecurityManager': 'nsIScriptSecurityManager',
    'search': 'nsIBrowserSearchService',
    'startup': 'nsIAppStartup',
    'storage': 'mozIStorageService',
    'strings': 'nsIStringBundleService',
    'sysinfo': 'nsIPropertyBag2',
    'telemetry': 'nsITelemetry',
    'tm': 'nsIThreadManager',
    'uriFixup': 'nsIURIFixup',
    'urlFormatter': 'nsIURLFormatter',
    'vc': 'nsIVersionComparator',
    'wm': 'nsIWindowMediator',
    'ww': 'nsIWindowWatcher',
}

for key, value in SERVICES.items():
    SERVICES[key] = {'value': partial(build_quick_xpcom,
                                      'getService', value)}

DANGEROUS_EVAL = {
    'err_id': ('javascript', 'dangerous_global', 'eval'),
    'description': ('Evaluation of strings as code can lead to security '
                    'vulnerabilities and performance issues, even in the '
                    'most innocuous of circumstances. Please avoid using '
                    '`eval` and the `Function` constructor when at all '
                    'possible.',
                    'Alternatives are available for most use cases. See '
                    'https://developer.mozilla.org/en-US/Add-ons/'
                    'Overlay_Extensions/XUL_School/'
                    'Appendix_C:_Avoid_using_eval_in_Add-ons '
                    'for more information.'),
    'signing_help':
        'Please try to avoid evaluating strings as code wherever possible. '
        'Read over the linked document for suggested alternatives. '
        'If you are referencing the `Function` constructor without calling '
        'it, and cannot avoid continuing to do so, consider alternatives '
        'such as calling `Object.getPrototypeOf` on an existing function '
        'object.',
    'signing_severity': 'high'}

FUNCTION_EXPORT_HELP = (
    'Given the potential security risks of exposing APIs to unprivileged '
    'code, extensions which use these APIs must undergo manual review for at '
    'least one submission. If you are not using these APIs to interact with '
    'content code, please consider alternatives, such as built-in '
    'message passing functionality.')

# GLOBAL_ENTITIES is also representative of the `window` object.
GLOBAL_ENTITIES = {
    u'window': {'value': lambda t: {'value': GLOBAL_ENTITIES}},
    u'null': {'literal': lambda t: JSWrapper(None, traverser=t)},
    u'Cc': {'readonly': False,
            'value':
                lambda t: GLOBAL_ENTITIES['Components']['value']['classes']},
    u'Ci': {'readonly': False,
            'value':
                lambda t: GLOBAL_ENTITIES['Components']['value']['interfaces']},
    u'Cu': {'readonly': False,
            'value':
                lambda t: GLOBAL_ENTITIES['Components']['value']['utils']},

    # From Services.jsm.
    u'Services': {'value': SERVICES},

    # From Preferences.jsm.
    # TODO: Support calls that return instances of this object which
    # operate on non-root branches.
    u'Preferences': {'value': {
        u'get': {'return': instanceactions.get_preference},
        u'reset': {'return': instanceactions.set_preference},
        u'resetBranch': {'return': instanceactions.set_preference},
        u'set': {'return': instanceactions.set_preference}}},

    u'AddonManager': {
        'readonly': False,
        'value': {
            u'autoUpdateDefault': {'readonly': SECURITY_PREF_MESSAGE},
            u'checkUpdateSecurity': {'readonly': SECURITY_PREF_MESSAGE},
            u'checkUpdateSecurityDefault': {'readonly': SECURITY_PREF_MESSAGE},
            u'updateEnabled': {'readonly': SECURITY_PREF_MESSAGE},
            u'getInstallForFile': ADDON_INSTALL_METHOD,
            u'getInstallForURL': ADDON_INSTALL_METHOD,
            u'installAddonsFromWebpage': ADDON_INSTALL_METHOD}},

    u'ctypes': {'dangerous': {
        'description': (
            'Insufficiently meticulous use of ctypes can lead to serious, '
            'and often exploitable, errors. The use of bundled binary code, '
            'or access to system libraries, may allow for add-ons to '
            'perform unsafe operations. All ctypes use must be carefully '
            'reviewed by a qualified reviewer.'),
        'editors_only': True,
        'signing_help': ('Please try to avoid interacting with or bundling '
                         'native binaries whenever possible. If you are '
                         'bundling binaries for performance reasons, please '
                         'consider alternatives such as Emscripten '
                         '(http://mzl.la/1KrSUh2), JavaScript typed arrays '
                         '(http://mzl.la/1Iw02sr), and Worker threads '
                         '(http://mzl.la/1OGfAcc).',
                         'Any code which makes use of the `ctypes` API '
                         'must undergo manual code review for at least one '
                         'submission.'),
        'signing_severity': 'high'}},

    u'document':
        {'value':
             {u'title':
                  {'overwriteable': True,
                   'readonly': False},
              u'defaultView':
                  {'value': lambda t: {'value': GLOBAL_ENTITIES}},
              u'loadOverlay':
                  {'dangerous':
                       lambda a, t, e:
                           not a or not _get_as_str(t(a[0])).lower()
                               .startswith(('chrome:', 'resource:'))},
              u'write': entity('document.write'),
              u'writeln': entity('document.write')}},

    # The nefariuos timeout brothers!
    u'setTimeout': {'dangerous': actions._call_settimeout},
    u'setInterval': {'dangerous': actions._call_settimeout},

    u'require': {'dangerous': actions._call_require},

    u'encodeURI': {'readonly': True},
    u'decodeURI': {'readonly': True},
    u'encodeURIComponent': {'readonly': True},
    u'decodeURIComponent': {'readonly': True},
    u'escape': {'readonly': True},
    u'unescape': {'readonly': True},
    u'isFinite': {'readonly': True},
    u'isNaN': {'readonly': True},
    u'parseFloat': {'readonly': True},
    u'parseInt': {'readonly': True},

    u'eval': {'dangerous': DANGEROUS_EVAL},
    u'Function': {'dangerous': DANGEROUS_EVAL},

    u'Object':
        {'value':
             {u'prototype': {'readonly': is_shared_scope},
              u'constructor':  # Just an experiment for now
                  {'value': lambda t: GLOBAL_ENTITIES['Function']}}},
    u'String':
        {'value':
             {u'prototype': {'readonly': is_shared_scope}},
         'return': call_definitions.string_global},
    u'Array':
        {'value':
             {u'prototype': {'readonly': is_shared_scope}},
         'return': call_definitions.array_global},
    u'Number':
        {'value':
             {u'prototype':
                  {'readonly': is_shared_scope},
              u'POSITIVE_INFINITY':
                  {'value': lambda t: JSWrapper(float('inf'), traverser=t)},
              u'NEGATIVE_INFINITY':
                  {'value': lambda t: JSWrapper(float('-inf'), traverser=t)}},
         'return': call_definitions.number_global},
    u'Boolean':
        {'value':
             {u'prototype': {'readonly': is_shared_scope}},
         'return': call_definitions.boolean_global},
    u'RegExp': {'value': {u'prototype': {'readonly': is_shared_scope}}},
    u'Date': {'value': {u'prototype': {'readonly': is_shared_scope}}},
    u'File': {'value': {u'prototype': {'readonly': is_shared_scope}}},

    u'Math':
        {'value':
             {u'PI':
                  {'value': lambda t: JSWrapper(math.pi, traverser=t)},
              u'E':
                  {'value': lambda t: JSWrapper(math.e, traverser=t)},
              u'LN2':
                  {'value': lambda t: JSWrapper(math.log(2), traverser=t)},
              u'LN10':
                  {'value': lambda t: JSWrapper(math.log(10), traverser=t)},
              u'LOG2E':
                  {'value': lambda t: JSWrapper(math.log(math.e, 2),
                                                traverser=t)},
              u'LOG10E':
                  {'value': lambda t: JSWrapper(math.log10(math.e),
                                                traverser=t)},
              u'SQRT2':
                  {'value': lambda t: JSWrapper(math.sqrt(2), traverser=t)},
              u'SQRT1_2':
                  {'value': lambda t: JSWrapper(math.sqrt(1/2), traverser=t)},
              u'abs':
                  {'return': python_wrap(abs, [('num', 0)])},
              u'acos':
                  {'return': python_wrap(math.acos, [('num', 0)])},
              u'asin':
                  {'return': python_wrap(math.asin, [('num', 0)])},
              u'atan':
                  {'return': python_wrap(math.atan, [('num', 0)])},
              u'atan2':
                  {'return': python_wrap(math.atan2, [('num', 0),
                                                      ('num', 1)])},
              u'ceil':
                  {'return': python_wrap(math.ceil, [('num', 0)])},
              u'cos':
                  {'return': python_wrap(math.cos, [('num', 0)])},
              u'exp':
                  {'return': python_wrap(math.exp, [('num', 0)])},
              u'floor':
                  {'return': python_wrap(math.floor, [('num', 0)])},
              u'log':
                  {'return': call_definitions.math_log},
              u'max':
                  {'return': python_wrap(max, [('num', 0)], nargs=True)},
              u'min':
                  {'return': python_wrap(min, [('num', 0)], nargs=True)},
              u'pow':
                  {'return': python_wrap(math.pow, [('num', 0),
                                                    ('num', 0)])},
              u'random': # Random always returns 0.5 in our fantasy land.
                  {'return': call_definitions.math_random},
              u'round':
                  {'return': call_definitions.math_round},
              u'sin':
                  {'return': python_wrap(math.sin, [('num', 0)])},
              u'sqrt':
                  {'return': python_wrap(math.sqrt, [('num', 1)])},
              u'tan':
                  {'return': python_wrap(math.tan, [('num', 0)])},
                  }},

    u'netscape':
        {'value':
             {u'security':
                  {'value':
                       {u'PrivilegeManager':
                            {'value':
                                 {u'enablePrivilege':
                                      {'dangerous': {
                                          'signing_help':
                                            'Any references to this API must '
                                            'be removed from your extension. '
                                            'Add-ons using this API will not '
                                            'be accepted for signing.',
                                          'signing_severity': 'high',
                                          'description': (
                                            'enablePrivilege is extremely '
                                            'dangerous, and nearly always '
                                            'unnecessary. It should not be '
                                            'used under any circumstances.'),
                                      }}}}}}}},
    u'navigator':
        {'value': {u'wifi': {'dangerous': True},
                   u'geolocation': {'dangerous': True}}},

    u'Components':
        {'dangerous_on_read':
             lambda t, e: bool(e.metadata.get('is_jetpack')),
         'value':
             {u'classes':
                  {'xpcom_wildcard': True,
                   'value':
                       {u'createInstance':
                           {'return': xpcom_const('createInstance')},
                        u'getService':
                           {'return': xpcom_const('getService')}}},
              'utils':
                  {'value': {u'evalInSandbox':
                                 {'dangerous': {
                                     'editors_only': 'true',
                                     'signing_help':
                                     DANGEROUS_EVAL['signing_help'],
                                     'signing_severity': 'low'}},
                             u'cloneInto':
                                 {'dangerous': {
                                     'editors_only': True,
                                     'signing_help': FUNCTION_EXPORT_HELP,
                                     'signing_severity': 'low',
                                     'description': (
                                         'Can be used to expose privileged '
                                         'functionality to unprivileged scopes. '
                                         'Care should be taken to ensure that '
                                         'this is done safely.')}},
                             u'exportFunction':
                                 {'dangerous': {
                                     'editors_only': True,
                                     'signing_help': FUNCTION_EXPORT_HELP,
                                     'signing_severity': 'low',
                                     'description': (
                                         'Can be used to expose privileged '
                                         'functionality to unprivileged scopes. '
                                         'Care should be taken to ensure that '
                                         'this is done safely.')}},
                             u'import':
                                 {'dangerous':
                                      lambda a, t, e:
                                          a and 'ctypes.jsm' in _get_as_str(t(a[0]))},

                             u'waiveXrays':
                                 {'return': call_definitions.js_unwrap}}},
              u'interfaces': {'value': INTERFACE_ENTITIES}}},
    u'extensions': {'dangerous': True},
    u'xpcnativewrappers': {'dangerous': True},

    u'XMLHttpRequest':
        {'value':
             {u'open':
                  {'dangerous':
                       # Ban synchronous XHR by making sure the third arg
                       # is absent and false.
                       lambda a, t, e:
                           a and len(a) >= 3 and
                           not t(a[2]).get_literal_value() and
                           'Synchronous HTTP requests can cause serious UI '
                           'performance problems, especially to users with '
                           'slow network connections.'}}},

    # Global properties are inherently read-only, though this formalizes it.
    u'Infinity':
        {'value':
             lambda t:
                 GLOBAL_ENTITIES[u'Number']['value'][u'POSITIVE_INFINITY']},
    u'NaN': {'readonly': True},
    u'undefined': {'readonly': True},

    u'innerHeight': {'readonly': False},
    u'innerWidth': {'readonly': False},
    u'width': {'readonly': False},
    u'height': {'readonly': False},
    u'top': {'readonly': actions._readonly_top},
    u'mozRequestAnimationFrame': {
        'value': actions._renamed_mozRequestAnimationFrame},

    u'content':
        {'context': 'content',
         'value':
             {u'document':
                  {'value': lambda t: GLOBAL_ENTITIES[u'document']}}},
    u'contentWindow':
        {'context': 'content',
         'value':
             lambda t: {'value': GLOBAL_ENTITIES}},
    u'_content': {'value': lambda t: GLOBAL_ENTITIES[u'content']},
    u'gBrowser':
        {'value':
             {u'contentDocument':
                  {'context': 'content',
                   'value': lambda t: CONTENT_DOCUMENT},
              u'contentWindow':
                  {'value':
                       lambda t: {'value': GLOBAL_ENTITIES}},
              u'selectedTab':
                  {'readonly': False}}},
    u'opener':
        {'value':
             lambda t: {'value': GLOBAL_ENTITIES}},

    u'XPCNativeWrapper':
        {'value':
             {u'unwrap':
                  {'return': call_definitions.js_unwrap}},
         'return': call_definitions.js_wrap},

    # Preference creation in pref defaults files
    u'pref': {'dangerous': actions._call_create_pref},
    u'user_pref': {'dangerous': actions._call_create_pref},

    u'unsafeWindow': {'dangerous': 'The use of unsafeWindow is insecure and '
                                   'should be avoided whenever possible. '
                                   'Consider using a different API if it is '
                                   'available in order to achieve similar '
                                   'functionality.'},

    u'XPCOMUtils':
        {'value': {u'categoryManager': {'value': CATEGORY_MANAGER}}},
    u'updateCharsetPopupMenu': entity('updateCharsetPopupMenu'),
    u'EditorSetDocumentCharacterSet': entity('EditorSetDocumentCharacterSet'),
    u'DisablePhishingWarning': entity('DisablePhishingWarning'),
    u'RoomInfo': entity('RoomInfo'),
    u'FillInHTMLTooltip': entity('FillInHTMLTooltip'),
    u'escapeXMLchars': entity('escapeXMLchars'),
    u'getNonHtmlRecipients': entity('getNonHtmlRecipients'),
    u'updateCharsetPopupMenu': entity('updateCharsetPopupMenu'),
    u'EditorSetDocumentCharacterSet': entity('EditorSetDocumentCharacterSet'),
    u'awArrowHit': entity('awArrowHit'),
    u'UpdateMailEditCharset': entity('UpdateMailEditCharset'),
    u'InitCharsetMenuCheckMark': entity('InitCharsetMenuCheckMark'),
    u'allowRemoteContentForSender': entity('allowRemoteContentForSender'),
    u'allowRemoteContentForSite': entity('allowRemoteContentForSite'),
    u'createNewHeaderView': entity('createNewHeaderView'),


    u'MarionetteComponent': {'dangerous_on_read': MARIONETTE_MESSAGE},
    u'MarionetteServer': {'dangerous_on_read': MARIONETTE_MESSAGE},

    'Application': {'dangerous_on_read': fuel_error},
    'NewTabURL': {'value': {'override': entity('NewTabURL.override')}},

    # Common third-party libraries
    'Handlebars': {
        'value': {
            'SafeString':
                {'dangerous':
                    UNSAFE_TEMPLATE_METHOD % 'Handlebars.SafeString'}}},
    # Angular
    '$sce': {
        'value': {
            'trustAs': {'dangerous':
                            UNSAFE_TEMPLATE_METHOD % '$sce.trustAs'},
            'trustAsHTML': {'dangerous':
                                UNSAFE_TEMPLATE_METHOD % '$sce.trustAsHTML'}}},
}

CONTENT_DOCUMENT = GLOBAL_ENTITIES[u'content']['value'][u'document']



from helper import CompatTestCase
from validator.compat import TB30_DEFINITION


class TestTB30Compat(CompatTestCase):
    """Test that compatibility tests for Thunderbird 30 are properly executed."""

    VERSION = TB30_DEFINITION

    def test_nsIMsgCompose_checkAndPopulateRecipients(self):
        for method in self.run_xpcom_for_compat(
                'nsIMsgCompose', ['checkAndPopulateRecipients']):
            self.assert_silent()
            self.assert_compat_error(type_='warning')

    def test_getNonHtmlRecipients(self):
        """Test that these patterns are flagged in Thunderbird 30."""
        self.run_script_for_compat('getNonHtmlRecipients()')
        self.assert_silent()
        self.assert_compat_error(type_='warning')

    def test_recentfolders_label(self):
        self.run_regex_for_compat('recentfolders.label')
        self.assert_silent()
        self.assert_compat_error(type_='warning')



from helper import CompatTestCase
from validator.compat import FX40_DEFINITION


class TestFX40Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 40 are properly executed."""

    VERSION = FX40_DEFINITION

    def test_setKeywordForBookmark(self):
        self.run_script_for_compat("""
            setKeywordForBookmark(25, 'foo');
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_getKeywordForBookmark(self):
        self.run_script_for_compat("""
            getKeywordForBookmark(25);
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_getURIForKeyword(self):
        self.run_script_for_compat("""
            getURIForKeyword('foo');
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_fuelIApplication(self):
        self.run_script_for_compat("""
            var Application = Components
                .classes["@mozilla.org/fuel/application;1"]
                .getService(Components.interfaces.fuelIApplication);
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_extIApplication(self):
        self.run_script_for_compat("""
            var Application = Components
                .classes["@mozilla.org/fuel/application;1"]
                .getService(Components.interfaces.extIApplication);
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_Application(self):
        self.run_script_for_compat("""
            Application.restart();
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_Dict_jsm(self):
        self.run_script_for_compat("""
            Components.utils.import("resource://gre/modules/Dict.jsm");
            var newDict = new Dict();
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_sessionstore_state_write_observer(self):
        self.run_script_for_compat("""
            let observer = {
                observe: function(subject, topic, data) {
                    if (topic === 'sessionstore-state-write') {
                        alert('this will not work anymore');
                    }
                }
            };
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_nsISSLErrorListener_observer(self):
        self.run_script_for_compat("""
            // I don't know how this is used...
            'nsISSLErrorListener';
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_sdk_widget_double(self):
        self.run_script_for_compat("""
            require("sdk/widget").Widget({
                id: "mozilla-icon",
                label: "My Mozilla Widget",
                contentURL: "http://www.mozilla.org/favicon.ico"
            });
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_sdk_widget_single(self):
        self.run_script_for_compat("""
            require('sdk/widget').Widget({
                id: "mozilla-icon",
                label: "My Mozilla Widget",
                contentURL: "http://www.mozilla.org/favicon.ico"
            });
        """)
        self.assert_silent()
        self.assert_compat_error()



from fnmatch import fnmatch as fnm

from validator.constants import (FF4_MIN, FIREFOX_GUID, FENNEC_GUID,
                                 THUNDERBIRD_GUID as TB_GUID, ANDROID_GUID,
                                 PACKAGE_DICTIONARY, )
import validator.decorator as decorator
from validator.decorator import version_range

# Detect blacklisted files based on their extension.
blacklisted_extensions = ('dll', 'exe', 'dylib', 'so',
                          'sh', 'class', 'swf')

blacklisted_magic_numbers = (
        (0x4d, 0x5a),  # EXE/DLL
        (0x5a, 0x4d),  # Alternative for EXE/DLL
        (0x7f, 0x45, 0x4c, 0x46),  # UNIX elf
        (0x23, 0x21),  # Shebang (shell script)
        (0xca, 0xfe, 0xba, 0xbe),  # Java + Mach-O (dylib)
        (0xca, 0xfe, 0xd0, 0x0d),  # Java (packed)
        (0xfe, 0xed, 0xfa, 0xce),  # Mach-O
        (0x46, 0x57, 0x53),  # Uncompressed SWF
        (0x43, 0x57, 0x53),  # ZLIB compressed SWF
)

# If there are more than 10 .class files in a package, it is flagged as a Java
# archive file.
JAVA_JAR_THRESHOLD = 10


def test_unknown_file(err, filename):
    'Tests some sketchy files that require silly code.'

    # Extract the file path and the name from the filename
    path = filename.split('/')
    name = path.pop()

    if name == 'chromelist.txt':
        err.notice(('testcases_packagelayout',
                    'test_unknown_file',
                    'deprecated_file'),
                   'Extension contains a deprecated file',
                   "The file \"%s\" is no longer supported by any modern "
                   'Mozilla product.' % filename,
                   filename)
        return True


@decorator.register_test(tier=1)
def test_blacklisted_files(err, xpi_package=None):
    'Detects blacklisted files and extensions.'

    flagged_files = []

    HELP = ('Please try to avoid interacting with or bundling '
            'native binaries whenever possible. If you are '
            'bundling binaries for performance reasons, please '
            'consider alternatives such as Emscripten '
            '(http://mzl.la/1KrSUh2), JavaScript typed arrays '
            '(http://mzl.la/1Iw02sr), and Worker threads '
            '(http://mzl.la/1OGfAcc).',
            'Any code which bundles native binaries must '
            'submit full source code, and undergo manual code review '
            'for at least one submission.')

    for name in xpi_package:
        file_ = xpi_package.info(name)
        # Simple test to ensure that the extension isn't blacklisted
        extension = file_['extension']
        if extension in blacklisted_extensions:
            # Note that there is a binary extension in the metadata
            err.metadata['contains_binary_extension'] = True
            flagged_files.append(name)
            continue

        # Perform a deep inspection to detect magic numbers for known binary
        # and executable file types.
        zip = xpi_package.zf.open(name)
        bytes = tuple([ord(x) for x in zip.read(4)])  # Longest is 4 bytes
        if [x for x in blacklisted_magic_numbers if bytes[0:len(x)] == x]:
            # Note that there is binary content in the metadata
            err.metadata['contains_binary_content'] = True
            err.warning(
                err_id=('testcases_packagelayout', 'test_blacklisted_files',
                        'disallowed_file_type'),
                warning='Flagged file type found',
                description='A file was found to contain flagged content '
                            '(i.e.: executable data, potentially '
                            'unauthorized scripts, etc.).',
                editors_only=True,
                signing_help=HELP,
                signing_severity='high',
                filename=name)

    if flagged_files:
        # Detect Java JAR files:
        if (sum(1 for f in flagged_files if f.endswith('.class')) >
                JAVA_JAR_THRESHOLD):
            err.warning(
                err_id=('testcases_packagelayout', 'test_blacklisted_files',
                        'java_jar'),
                warning='Java JAR file detected.',
                description='A Java JAR file was detected in the add-on.',
                filename=xpi_package.filename,
                signing_help=HELP,
                signing_severity='high')
        else:
            err.warning(
                err_id=('testcases_packagelayout',
                        'test_blacklisted_files',
                        'disallowed_extension'),
                warning='Flagged file extensions found.',
                description=('Files whose names end with flagged extensions '
                             'have been found in the add-on.',
                             'The extension of these files are flagged '
                             'because they usually identify binary '
                             'components. Please see '
                             'http://addons.mozilla.org/developers/docs/'
                             'policies/reviews#section-binary'
                             ' for more information on the binary content '
                             'review process.',
                             '\n'.join(flagged_files)),
                editors_only=True,
                signing_help=HELP,
                signing_severity='medium',
                filename=name)


@decorator.register_test(tier=1)
def test_godlikea(err, xpi_package):
    """Test to make sure that the godlikea namespace is not in use."""

    if 'chrome/godlikea.jar' in xpi_package:
        err.error(
            err_id=('testcases_packagelayout',
                    'test_godlikea'),
            error="Banned 'godlikea' chrome namespace",
            description="The 'godlikea' chrome namepsace is generated from a "
                        'template and should be replaced with something '
                        'unique to your add-on to avoid name conflicts.',
            filename='chrome/godlikea.jar')


@decorator.register_test(
        tier=5,
        versions={FIREFOX_GUID: version_range('firefox', FF4_MIN),
                  TB_GUID: version_range('thunderbird', '3.3a4pre'),
                  FENNEC_GUID: version_range('fennec', '4.0'),
                  ANDROID_GUID: version_range('android', '11.0a1')})
def test_compatibility_binary(err, xpi_package):
    """
    Flags only binary content as being incompatible with future app releases.
    """

    description = ('Add-ons with binary components must have their '
                   'compatibility manually adjusted. Please test your add-on '
                   'against the new version before updating your maxVersion.')

    chrome = err.get_resource('chrome.manifest')
    if not chrome:
        return

    for triple in chrome.triples:
        if triple['subject'] == 'binary-component':
            err.metadata['binary_components'] = True
            err.notice(
                err_id=('testcases_packagelayout',
                        'test_compatibility_binary',
                        'disallowed_file_type'),
                notice='Flagged file type found',
                description=('A file (%s) was registered as a binary '
                             'component. Binary files may not be submitted to '
                             'AMO unless accompanied by source code.'
                                % triple['predicate'],
                             description),
                filename=triple['predicate'],
                compatibility_type='error')


@decorator.register_test(tier=1)
def test_layout_all(err, xpi_package):
    'Tests the well-formedness of extensions.'

    # Subpackages don't need to be tested for a manifest.
    if xpi_package.subpackage:
        return

    if (not err.get_resource('has_install_rdf') and
            not err.get_resource('bad_install_rdf') and
            not err.get_resource('has_package_json') and
            not err.get_resource('has_manifest_json')):
        err.error(('testcases_packagelayout',
                  'test_layout_all',
                  'missing_install_rdf'),
                  'Add-on missing manifest.',
                  'All add-ons require a manifest file.')
        return

    package_namelist = list(xpi_package.zf.namelist())
    package_nameset = set(package_namelist)
    if len(package_namelist) != len(package_nameset):
        err.error(
            err_id=('testcases_packagelayout', 'test_layout_all',
                    'duplicate_entries'),
            error='Package contains duplicate entries',
            description='The package contains multiple entries with the same '
                        'name. This practice has been banned. Try unzipping '
                        'and re-zipping your add-on package and try again.')


# This test needs to happen in tier 2. install.rdf analysis happens in tier
# 1, which may cause this to generate false positives if it's also on tier 1
# (bug 631340)
@decorator.register_test(tier=2)
def test_emunpack(err, xpi_package):

    if err.get_resource('em:unpack') != 'true':
        # Covers bug 597255

        # Dictionaries should always be unpacked
        fails = err.detected_type == PACKAGE_DICTIONARY
        if not fails:
            executables = ('exe', 'dll', 'so', 'dylib', 'exe', 'bin')
            # Search for unpack-worthy files
            for file_ in xpi_package:
                if file_.startswith('chrome/icons/default/'):
                    fails = True
                    break
                # Executables in /components/ should also be flagged.
                if (file_.startswith('components/') and
                    [x for x in executables if
                     file_[-len(x) - 1:] == '.%s' % x]):
                    fails = True
                    break

        if fails:
            err.warning(('testcases_packagelayout',
                         'test_emunpack',
                         'should_be_true'),
                        'Add-on should set <em:unpack> to true',
                        'The add-on meets criteria indicating that it will '
                        'not work correctly in Gecko 4 and above. Set '
                        "<em:unpack> to 'true' in the install.rdf file to fix "
                        'this.',
                        filename='install.rdf',
                        tier=1)
            return

        # Covers bug 551714

        # This only applies to FF4
        if not err.get_resource('ff4'):
            return

        for file_ in xpi_package:
            if file_.endswith('.jar'):
                err.notice(
                    err_id=('packagelayout', 'emunpack', 'should_be_false'),
                    notice='Add-on contains JAR files, no <em:unpack>',
                    description='The add-on contains JAR files and does not '
                                "set <em:unpack> to 'true'. This can result "
                                'in performance issues in add-ons that target '
                                'Gecko 4 and above. You should either no '
                                'longer use JAR files to package your chrome '
                                'files (preferred) or add '
                                '`<em:unpack>true</em:unpack>` to the '
                                'install.rdf.',
                    filename='install.rdf',
                    tier=1)
                return


@decorator.register_test(tier=1, expected_type=3)
def test_dictionary_layout(err, xpi_package=None):
    """Ensures that dictionary packages contain the necessary
    components and that there are no other extraneous files lying
    around."""

    # Define rules for the structure.
    mandatory_files = [
        'install.rdf',
        'dictionaries/*.aff',
        'dictionaries/*.dic']
    whitelisted_files = [
        'install.js',
        'icon.png',
        'icon64.png',
        'dictionaries/*.aff',  # List again because there must >0
        'dictionaries/*.dic',
        'chrome.manifest',
        'chrome/*']
    whitelisted_extensions = ('txt',)

    test_layout(err, xpi_package, mandatory_files,
                whitelisted_files, whitelisted_extensions,
                'dictionary')


@decorator.register_test(tier=1, expected_type=4)
def test_langpack_layout(err, xpi_package=None):
    """Ensures that language packs only contain exactly what they
    need and nothing more. Otherwise, somebody could sneak something
    sneaking into them."""

    # Define rules for the structure.
    mandatory_files = [
        'install.rdf',
        'chrome.manifest']
    whitelisted_files = ['chrome/*.jar']
    whitelisted_extensions = ('manifest', 'rdf', 'jar', 'dtd',
                              'properties', 'xhtml', 'css')

    test_layout(err, xpi_package, mandatory_files,
                whitelisted_files, whitelisted_extensions,
                'language pack')


@decorator.register_test(tier=1, expected_type=2)
def test_theme_layout(err, xpi_package=None):
    """Ensures that themes only contain exactly what they need and
    nothing more. Otherwise, somebody could sneak something sneaking
    into them."""

    # Define rules for the structure.
    mandatory_files = [
        'install.rdf',
        'chrome.manifest']
    whitelisted_files = ['chrome/*.jar']

    test_layout(err, xpi_package, mandatory_files,
                whitelisted_files, None, 'theme')


def test_layout(err, xpi, mandatory, whitelisted,
                white_extensions=None, pack_type='Unknown Addon'):
    """Tests the layout of a package. Pass in the various types of files
    and their levels of requirement and this guy will figure out which
    files should and should not be in the package."""

    # A shortcut to prevent excessive lookups

    for file_ in xpi:

        if file_ == '.DS_Store' or file_.startswith('__MACOSX/'):
            continue

        # Remove the file from the mandatory file list.
        #if file_ in mandatory_files:
        mfile_list = [mf for mf in mandatory if fnm(file_, mf)]
        if mfile_list:
            # Isolate the mandatory file pattern and remove it.
            mfile = mfile_list[0]
            mandatory.remove(mfile)
            continue

        # Test if the file is in the whitelist.
        if any(fnm(file_, wlfile) for wlfile in whitelisted):
            continue

        # Is it a directory?
        if file_.endswith('/'):
            continue

        # Is the file a sketch file?
        if test_unknown_file(err, file_):
            continue

        # Is it a whitelisted file type?
        if white_extensions is None or file_.split('.')[-1] in white_extensions:
            continue

        # Otherwise, report an error.
        err.warning(('testcases_packagelayout',
                     'test_layout',
                     'unknown_file'),
                    'Unknown file found in add-on',
                    ['Files have been detected that are not allowed within '
                     'this type of add-on. Remove the file or use an '
                     'alternative, supported file format instead.',
                     'Detected file: %s' % file_],
                    file_)

    # If there's anything left over, it means there's files missing
    if mandatory:
        err.warning(('testcases_packagelayout',
                     'test_layout',
                     'missing_required'),
                    'Required file missing',
                    ['This add-on is missing required files. Consult the '
                     'documentation for a full list of required files.',
                     "Add-ons of type '%s' require files: %s" %
                          (pack_type, ', '.join(mandatory))])



import types
import instanceproperties

from validator.constants import JETPACK_URI_URL


class JSObject(object):
    """
    Mimics a JS object (function) and is capable of serving as an active
    context to enable static analysis of `with` statements.
    """

    def __init__(self, data=None, unwrapped=False):
        self.type_ = 'object'  # For use when an object is pushed as a context.
        self.data = {u'prototype': lambda: JSPrototype()}
        if data:
            self.data.update(data)
        self.messages = []
        self.is_unwrapped = unwrapped

        self.recursing = False

    def get(self, name, instantiate=False, traverser=None):
        'Returns the value associated with a property name'
        name = unicode(name)
        output = None

        if name == 'wrappedJSObject':
            if traverser:
                traverser._debug('Requested unwrapped JS object...')
            clone = JSObject(unwrapped=True)
            clone.data = self.data
            return JSWrapper(clone, traverser=traverser)

        if name in self.data:
            output = self.data[name]
            if callable(output):
                output = output()
        elif instantiate:
            output = JSWrapper(JSObject(), dirty=True, traverser=traverser)
            self.set(name, output, traverser=traverser)

        if traverser:
            modifier = instanceproperties.get_operation('get', name)
            if modifier:
                modifier(traverser)

        if output is None:
            return JSWrapper(JSObject(), dirty=True, traverser=traverser)
        if not isinstance(output, JSWrapper):
            output = JSWrapper(output, traverser=traverser)
        return output

    def get_literal_value(self):
        return u'[object Object]'

    def set(self, name, value, traverser=None):
        if traverser:
            modifier = instanceproperties.get_operation('set', name)
            if modifier:
                modified_value = modifier(value, traverser)
                if modified_value is not None:
                    value = modified_value

            if self.is_unwrapped:
                traverser.warning(
                    err_id=('testcases_javascript_jstypes', 'JSObject_set',
                            'unwrapped_js_object'),
                    warning="Assignment of unwrapped JS Object's properties.",
                    description='Improper use of unwrapped JS objects can '
                                'result in serious security vulnerabilities. '
                                'Please reconsider your use of unwrapped '
                                'JS objects.',
                    signing_help='Please avoid assigning to properties of '
                                 'unwrapped objects from unprivileged scopes, '
                                 'unless you are using an export API '
                                 '(http://mzl.la/1fvvgm9) to expose API '
                                 'functions to content scopes. '
                                 'In this case, however, please note that '
                                 'your add-on will be required to undergo '
                                 'manual code review for at least one '
                                 'submission.',
                    signing_severity='high')

        self.data[name] = value

    def has_var(self, name):
        name = unicode(name)
        return name in self.data

    def output(self):
        if self.recursing:
            return u'(recursion)'

        # Prevent unruly recursion with a recursion buster.
        self.recursing = True

        output_dict = {}
        for key in self.data.keys():
            if callable(self.data[key]):
                continue
            elif (isinstance(self.data[key], JSWrapper) and
                  self.data[key].value == self):
                output_dict[key] = u'(self)'
            elif self.data[key] is None:
                output_dict[key] = u'(None)'
            else:
                output_dict[key] = self.data[key].output()

        # Pop from the recursion buster.
        self.recursing = False

        output = []
        if self.is_unwrapped:
            output.append('<unwrapped>: ')
        output.append(str(output_dict))
        return ''.join(output)


class JSContext(JSObject):
    """A variable context"""

    def __init__(self, context_type, unwrapped=False):
        super(JSContext, self).__init__(unwrapped=unwrapped)
        self.type_ = context_type
        self.data = {}


class JSWrapper(object):
    """Wraps a JS value and handles contextual functions for it."""

    def __init__(self, value=None, const=False, dirty=False, lazy=False,
                 is_global=False, traverser=None, callable=False,
                 setter=None, context='chrome'):

        if is_global:
            assert not value

        if traverser is not None:
            traverser.debug_level += 1
            traverser._debug('-----New JSWrapper-----')
            if isinstance(value, JSWrapper):
                traverser._debug('>>> Rewrap <<<')
            traverser.debug_level -= 1

        self.const = const
        self.traverser = traverser
        self.value = None  # Instantiate the placeholder value
        self.is_global = False  # Not yet...
        self.dirty = False  # Also not yet...
        self.context = context

        # Used for predetermining set operations
        self.setter = setter

        if value is not None:
            self.set_value(value, overwrite_const=True)

        if not self.is_global:
            self.is_global = is_global  # Globals are built seperately

        self.dirty = dirty or self.dirty
        self.lazy = lazy
        self.callable = callable

        # This will be set in actions.py if needed.
        self.global_parent = False

    def set_value(self, value, traverser=None, overwrite_const=False):
        """Assigns a value to the wrapper"""

        # Use a global traverser if it's present.
        if traverser is None:
            traverser = self.traverser

        if self.const and not overwrite_const:
            traverser.err.warning(
                err_id=('testcases_javascript_traverser',
                        'JSWrapper_set_value', 'const_overwrite'),
                warning='Overwritten constant value',
                description='A variable declared as constant has been '
                            'overwritten in some JS code.',
                filename=traverser.filename,
                line=traverser.line,
                column=traverser.position,
                context=traverser.context)

        # Process any setter/modifier
        if self.setter:
            traverser._debug('Running setter on JSWrapper...')
            value = self.setter(value, traverser) or value or None

        if value == self.value:
            return self

        # We want to obey the permissions of global objects
        if (self.is_global and
            (not traverser or
             not (traverser.is_jsm or
                  traverser.err.get_resource('em:bootstrap') == 'true')) and
            (isinstance(self.value, dict) and
             not self.value.get('overwriteable', True))):
            traverser.err.warning(
                err_id=('testcases_javascript_jstypes', 'JSWrapper_set_value',
                        'global_overwrite'),
                warning='Global overwrite',
                description='An attempt to overwrite a global variable was '
                            'made in some JS code.',
                filename=traverser.filename,
                line=traverser.line,
                column=traverser.position,
                context=traverser.context)
            return self

        if isinstance(value, (bool, str, int, float, long, unicode)):
            self.inspect_literal(value)
            value = JSLiteral(value)
        # If the value being assigned is a wrapper as well, copy it in
        elif isinstance(value, JSWrapper):
            self.value = value.value
            self.lazy = value.lazy
            self.dirty = value.dirty
            self.is_global = value.is_global
            self.context = value.context
            # const does not carry over on reassignment
            return self
        elif callable(value):
            value = value(t=traverser)

        if not isinstance(value, dict):
            self.is_global = False
        elif 'context' in value:
            self.context = value['context']

        self.value = value
        return self

    def has_property(self, property):
        """Returns a boolean value representing the presence of a property"""
        if isinstance(self.value, JSLiteral):
            return False
        return isinstance(self.value, JSObject)

    def get(self, traverser, name, instantiate=False):
        """Retrieve a property from the variable."""

        value = self.value
        dirty = value is None
        context = self.context
        if self.is_global:
            if 'value' not in value:
                output = JSWrapper(JSObject(), traverser=traverser)
                output.value = {}

                def apply_value(name):
                    if name in self.value:
                        output.value[name] = self.value[name]

                map(apply_value, ('dangerous', 'readonly', 'context', 'name'))
                output.is_global = True
                output.context = self.context
                return output

            def _evaluate_lambdas(node):
                if callable(node):
                    return _evaluate_lambdas(node(t=traverser))
                else:
                    return node

            value_val = value['value']
            value_val = _evaluate_lambdas(value_val)

            if isinstance(value_val, dict):
                if name in value_val:
                    value_val = _evaluate_lambdas(value_val[name])
                    output = traverser._build_global(name=name,
                                                     entity=value_val)
                    if 'context' not in value_val:
                        output.context = self.context
                    return output
            else:
                value = value_val

        # Process any getters that are present for the current property.
        modifier = instanceproperties.get_operation('get', name)
        if modifier:
            modifier(traverser)

        if value is not None and issubclass(type(value), JSObject):
            output = value.get(name, instantiate=instantiate,
                               traverser=traverser)
        else:
            output = None

        if not isinstance(output, JSWrapper):
            output = JSWrapper(
                output, traverser=traverser, dirty=output is None or dirty)

        output.context = context

        # If we can predetermine the setter for the wrapper, we can save a ton
        # of lookbehinds in the future. This greatly simplifies the
        # MemberExpression support.
        setter = instanceproperties.get_operation('set', name)
        if setter:
            output.setter = setter
        return output

    def del_value(self, member):
        """The member `member` will be deleted from the value of the wrapper"""
        if self.is_global:
            self.traverser.err.warning(
                err_id=('testcases_js_jstypes', 'del_value',
                        'global_member_deletion'),
                warning='Global member deletion',
                description='Members of global object may not be deleted.',
                filename=self.traverser.filename,
                line=self.traverser.line,
                column=self.traverser.position,
                context=self.traverser.context)
            return
        elif isinstance(self.value, (JSObject, JSPrototype)):
            if member not in self.value.data:
                return
            del self.value.data[member]

    def contains(self, value):
        """Serves 'in' for BinaryOperators for lists and dictionaries"""

        # Unwrap the rvalue.
        if isinstance(value, JSWrapper):
            value = value.get_literal_value()

        if isinstance(self.value, JSArray):
            from actions import _get_as_num
            index = int(_get_as_num(value))
            if len(self.value.elements) > index >= 0:
                return True

        if isinstance(self.value, (JSArray, JSObject, JSPrototype)):
            return self.value.has_var(value)

        # Nothing else supports "in"
        return False

    def is_literal(self):
        """Returns whether the content is a literal"""
        return isinstance(self.value, JSLiteral)

    def get_literal_value(self):
        """Returns the literal value of the wrapper"""

        if self.is_global:
            if 'literal' in self.value:
                return self.value['literal'](self.traverser)
            else:
                return '[object Object]'
        if self.value is None:
            return None

        output = self.value.get_literal_value()
        return output

    def output(self):
        """Returns a readable version of the object"""
        if self.value is None:
            return '(None)'
        elif self.is_global:
            return '(Global)'

        return self.value.output()

    def inspect_literal(self, value):
        """
        Inspect the value of a literal to see whether it contains a flagged
        value.
        """

        # Don't do any processing if we can't return an error.
        if not self.traverser:
            return

        self.traverser._debug('INSPECTING: %s' % value)

        if isinstance(value, types.StringTypes):
            if ('is_jetpack' in self.traverser.err.metadata and
                    value.startswith('resource://') and
                    '-data/' in value):
                # Since Jetpack files are ignored, this should not be scanning
                # anything inside the jetpack directories.
                self.traverser.warning(
                    err_id=('javascript_js_jstypes', 'jswrapper',
                            'jetpack_abs_uri'),
                    warning='Absolute URIs in Jetpack 1.4 are disallowed',
                    description=('As of Jetpack 1.4, absolute URIs are no '
                                 'longer allowed within add-ons.',
                                 'See %s for more information.'
                                 % JETPACK_URI_URL),
                    compatibility_type='error')

    def __unicode__(self):
        """Returns a textual version of the object."""
        return unicode(self.get_literal_value())


class JSLiteral(JSObject):
    """Represents a literal JavaScript value."""

    def __init__(self, value=None, unwrapped=False):
        super(JSLiteral, self).__init__(unwrapped=unwrapped)
        self.value = value
        self.source = None

    def set_value(self, value):
        self.value = value

    def __str__(self):
        if isinstance(self.value, bool):
            return str(self.output()).lower()
        return str(self.output())

    def output(self):
        return self.value

    def get_literal_value(self):
        'Returns the literal value of a this literal. Heh.'
        return self.value


class JSPrototype(JSObject):
    """
    A lazy JavaScript object that is assumed not to contain any default
    methods.
    """

    def __init__(self, unwrapped=False):
        super(JSPrototype, self).__init__(unwrapped=unwrapped)
        self.data = {}

    def get(self, name, instantiate=False, traverser=None):
        name = unicode(name)
        output = super(JSPrototype, self).get(name, instantiate, traverser)
        if output is not None:
            return output
        if name == 'prototype':
            prototype = JSWrapper(JSPrototype(), traverser=traverser)
            self.data[name] = prototype

        return output


class JSArray(JSObject):
    """A class that represents both a JS Array and a JS list."""

    def __init__(self, unwrapped=False):
        super(JSArray, self).__init__(unwrapped=unwrapped)
        self.elements = []

    def get(self, index, instantiate=False, traverser=None):
        if index == 'length':
            return len(self.elements)

        # Courtesy of Ian Bicking: http://bit.ly/hxv6qt
        try:
            output = self.elements[int(index.strip().split()[0])]
            if not isinstance(output, JSWrapper):
                output = JSWrapper(output, traverser=traverser)
            return output
        except (ValueError, IndexError, KeyError):
            return super(JSArray, self).get(index, instantiate, traverser)

    def get_literal_value(self):
        """Arrays return a comma-delimited version of themselves."""

        if self.recursing:
            return u'(recursion)'

        self.recursing = True

        # Interestingly enough, this allows for things like:
        # x = [4]
        # y = x * 3 // y = 12 since x equals "4"

        output = u','.join(
            [unicode(w.get_literal_value() if w is not None else u'') for w in
             self.elements if
             not (isinstance(w, JSWrapper) and w.value == self)])

        self.recursing = False
        return output

    def set(self, index, value, traverser=None):
        try:
            index = int(index)
            # Ignore floating point indexes
            if index != float(index) or index < 0:
                return super(JSArray, self).set(value, traverser)
        except ValueError:
            return super(JSArray, self).set(index, value, traverser)

        if len(self.elements) > index:
            self.elements[index] = JSWrapper(value=value, traverser=traverser)
        else:
            # Max out the array size at 100000
            index = min(index, 100000)
            # Assigning to an index higher than the top of the list pads the
            # list with nulls
            while len(self.elements) < index:
                self.elements.append(None)
            self.elements.append(JSWrapper(value=value, traverser=traverser))

    def output(self):
        return u'[%s]' % self.get_literal_value()



import re
from functools import wraps

from validator.compat import (
    FX38_DEFINITION, FX39_DEFINITION, FX40_DEFINITION, FX41_DEFINITION,
    FX43_DEFINITION, FX44_DEFINITION, FX45_DEFINITION, FX46_DEFINITION,
    TB29_DEFINITION, TB30_DEFINITION, TB31_DEFINITION)
from validator.constants import BUGZILLA_BUG, MDN_DOC
from validator.contextgenerator import ContextGenerator
from .chromemanifest import DANGEROUS_CATEGORIES, DANGEROUS_CATEGORY_WARNING
from .javascript.predefinedentities import (BANNED_PREF_BRANCHES,
                                            BANNED_PREF_REGEXPS,
                                            MARIONETTE_MESSAGE)


registered_regex_tests = []


NP_WARNING = 'Network preferences may not be modified.'
EUP_WARNING = 'Extension update settings may not be modified.'
NSINHS_LINK = MDN_DOC % 'XPCOM_Interface_Reference/nsINavHistoryService'


def run_regex_tests(document, err, filename, context=None, is_js=False,
                    explicit=False):
    """Run all of the regex-based JS tests."""

    # When `explicit` is true, only run tests which explicitly
    # specify which files they're applicable to.

    if context is None:
        context = ContextGenerator(document)

    # Run all of the registered tests.
    for cls in registered_regex_tests:
        if not hasattr(cls, 'applicable'):
            if explicit:
                continue
        else:
            if not cls.applicable(err, filename, document):
                continue

        t = cls(err, document, filename, context)
        # Run standard tests.
        for test in t.tests():
            test()
        # Run tests that would otherwise be guarded by `is_js`.
        if is_js:
            for test in t.js_tests():
                test()


def register_generator(cls):
    registered_regex_tests.append(cls)
    return cls


def merge_description(base, description):
    base = base.copy()
    if isinstance(description, dict):
        base.update(description)
    else:
        base['description'] = description
    return base


class RegexTestGenerator(object):
    """
    This stubs out a test generator object. By decorating inheritors of this
    class with `@register_generator`, the regex tests will be run on any files
    that are passed to `run_regex_tests()`.
    """

    def __init__(self, err, document, filename, context):
        self.err = err
        self.document = document
        self.filename = filename
        self.context = context

        self.app_versions_fallback = None

    def tests(self):
        """Override this with a generator that produces the regex tests."""
        return []

    def js_tests(self):
        """
        Override this with a generator that produces regex tests that are
        exclusive to JavaScript.
        """
        return []

    def get_test(self, pattern, title, message, log_function=None,
                 compat_type=None, app_versions=None, flags=0):
        """
        Return a function that, when called, will log a compatibility warning
        or error.
        """
        app_versions = app_versions or self.app_versions_fallback
        log_function = log_function or self.err.warning
        @wraps(log_function)
        def wrapper():
            matched = False
            for match in re.finditer(pattern, self.document, flags):
                kw = {'err_id': ('testcases_regex', 'generic', '_generated'),
                      log_function.__name__: title,
                      'filename': self.filename,
                      'line': self.context.get_line(match.start()),
                      'context': self.context,
                      'compatibility_type': compat_type,
                      'for_appversions': app_versions,
                      'tier': self.err.tier if app_versions is None else 5}

                log_function(**merge_description(kw, message))
                matched = True
            return matched

        return wrapper

    def get_test_bug(self, bug, pattern, title, message, **kwargs):
        """Helper function to mix in a bug number."""
        message = [message,
                   'See bug %s for more information.' % BUGZILLA_BUG % bug]
        return self.get_test(pattern, title, message, **kwargs)


@register_generator
class GenericRegexTests(RegexTestGenerator):
    """Test for generic, banned patterns in a document being scanned."""

    def tests(self):
        # globalStorage.(.+)password test removed for bug 752740

        yield self.get_test(
                r'resource://services-sync',
                'Sync services objects are not intended to be re-used',
                'The Sync services objects are not intended to be re-used, and '
                'they often change in ways that break add-ons. It is strongly '
                'recommended that you do not rely on them.')

    def js_tests(self):
        yield self.get_test(
                'mouse(move|over|out)',
                'Mouse events may cause performance issues.',
                'The use of `mousemove`, `mouseover`, and `mouseout` is '
                'discouraged. These events are dispatched with high frequency '
                'and can cause severe performance issues.',
                log_function=self.err.warning)


@register_generator
class DOMMutationRegexTests(RegexTestGenerator):
    """
    These regex tests will test that DOM mutation events are not used. These
    events have extreme performance penalties associated with them and are
    currently deprecated.

    Added from bug 642153
    """

    EVENTS = ('DOMAttrModified', 'DOMAttributeNameChanged',
              'DOMCharacterDataModified', 'DOMElementNameChanged',
              'DOMNodeInserted', 'DOMNodeInsertedIntoDocument',
              'DOMNodeRemoved', 'DOMNodeRemovedFromDocument',
              'DOMSubtreeModified', )

    def js_tests(self):
        for event in self.EVENTS:
            yield self.get_test(
                    event,
                    'DOM mutation event use prohibited',
                    'DOM mutation events are flagged because of their '
                    'deprecated status as well as their extreme inefficiency. '
                    'Consider using a different event.')


@register_generator
class NewTabRegexTests(RegexTestGenerator):
    """
    Attempts to code using roundabout methods of overriding the new tab page.
    """

    PATTERN = r"""
        (?x)
        == \s* ["']about:(newtab|blank)["'] |
        ["']about:(newtab|blank)["'] \s* == |
        /\^?about:newtab\$?/ \s* \. test\b |
        \?what=newtab
    """

    def tests(self):
        yield self.get_test(
            self.PATTERN,
            'Possible attempt to override new tab page',
            {'description': (
                'The new tab page should be changed only by writing '
                'to the appropriate preference in the default preferences '
                'branch. Such changes may only be made after an explicit '
                'user opt-in, unless the add-on was explicitly and directly '
                'installed by the user, and changing the new tab page is its '
                'primary purpose.',
                'If this code does not change the behavior of the new tab '
                'page, it may be ignored.'),
             'signing_help':
                'Extensions may not programmatically override the new tab '
                'page. If this code has another purpose, we nonetheless '
                'recommend against testing URLs for these values, since '
                'results can be unpredictable, and better options usually '
                'exist. If you cannot avoid making these tests, please leave '
                'this code unchanged, and it will be ignored in future '
                'submissions.',
             'signing_severity': 'low'})


@register_generator
class UnsafeTemplateRegexTests(RegexTestGenerator):
    """
    Checks for the use of unsafe template escape sequences.
    """

    @classmethod
    def applicable(cls, err, filename, document):
        # Perhaps this should just run on all non-binary files, but
        # we'll try to be more selective for the moment.
        return bool(re.match(r'.*\.(js(|m)|hbs|handlebars|mustache|'
                             r'(t|)htm(|l)|tm?pl)',
                             filename))

    def tests(self):
        for unsafe, safe in (('<%=', '<%-'),
                             ('{{{', '{{'),
                             ('ng-bind-html-unsafe=', 'ng-bind-html')):
            yield self.get_test(
                    re.escape(unsafe),
                    'Potentially unsafe template escape sequence',
                    'The use of non-HTML-escaping template escape sequences is '
                    'potentially dangerous and highly discouraged. Non-escaped '
                    'HTML may only be used when properly sanitized, and in most '
                    'cases safer escape sequences such as `%s` must be used '
                    'instead.' % safe)


@register_generator
class JSPrototypeExtRegexTests(RegexTestGenerator):
    """
    These regex tests will ensure that the developer is not modifying the
    prototypes of the global JS types.

    Added from bug 696172
    """

    @classmethod
    def applicable(cls, err, filename, document):
        return not (filename.endswith('.jsm') or 'EXPORTED_SYMBOLS' in document)

    def js_tests(self):
        yield self.get_test(
                r'\b(String|Object|Number|Date|RegExp|Function|Boolean|Array|'
                r'Iterator)\.prototype(\.[a-zA-Z0-9]+|\[.+\])\s*=',
                'JS prototype extension',
                'It appears that an extension of a built-in JS type was made. '
                'This is not allowed for security and compatibility reasons.',
                flags=re.I)


@register_generator
class XPIProviderRegexTests(RegexTestGenerator):
    """
    These regex tests ensure that developers are not using the XPIProvider
    or AddonManagerInternal symbols.

    Added from bug 1200929.
    """

    def tests(self):
        msg = 'Access to AddonManagerInternal and XPIProvider is not allowed.'

        yield self.get_test(
            r'\b(XPIProvider|AddonManagerInternal)\b',
            msg, {'description': msg, 'signing_severity': 'high'})


class CompatRegexTestHelper(RegexTestGenerator):
    """
    A helper that makes it easier to stay DRY. This will automatically check
    for applicability against the value set as the app_versions_fallback.
    """

    def __init__(self, *args, **kwargs):
        super(CompatRegexTestHelper, self).__init__(*args, **kwargs)
        self.app_versions_fallback = self.VERSION

    @classmethod
    def applicable(cls, err, filename, document):
        return err.supports_version(cls.VERSION)


DEP_INTERFACE = 'Deprecated interface in use'
DEP_INTERFACE_DESC = ('This add-on uses `%s`, which is deprecated in Gecko %d '
                      'and should not be used.')


@register_generator
class Gecko38RegexTests(CompatRegexTestHelper):
    """Regex tests for Gecko 38 updates."""

    VERSION = FX38_DEFINITION

    def tests(self):
        yield self.get_test(
            r'\bmozIndexedDB\b',
            'mozIndexedDB has been removed.',
            'mozIndexedDB has been removed. You should use indexedDB instead. '
            'See %s for more information.' % BUGZILLA_BUG % 975699,
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\b(nsICompositionStringSynthesizer|'
            r'sendCompositionEvent|'
            r'createCompositionStringSynthesizer)\b',
            'nsICompositionStringSynthesizer, sendCompositionEvent and '
            'createCompositionStringSynthesizer were removed.',
            'The nsICompositionStringSynthesizer interface and the '
            'sendCompositionEvent and createCompositionStringSynthesizer '
            'functions have been removed. See %s for more information.'
            % MDN_DOC
            % 'Mozilla/Tech/XPCOM/Reference/Interface/nsITextInputProcessor',
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\b(newChannel2|asyncFetch2)\b',
            'asyncFetch2 and newChannel2 are now deprecated.',
            'asyncFetch2 and newChannel2 are now deprecated. Use asyncFetch '
            'or newChannel instead. See %s for more information.'
            % BUGZILLA_BUG % 1125618,
            compat_type='warning')

        yield self.get_test(
            r'\b(onProxyAvailable|asyncResolve)\b',
            'The onProxyAvailable and asyncResolve functions have changed.',
            'The onProxyAvailable and asyncResolve functions have changed. '
            'They now take an nsIChannel instead of an nsIURI as an argument. '
            'See %s for more information.' % BUGZILLA_BUG % 436344,
            log_function=self.err.warning,
            compat_type='error')


@register_generator
class Gecko39RegexTests(CompatRegexTestHelper):
    """Regex tests for Gecko 39 updates."""

    VERSION = FX39_DEFINITION

    def tests(self):
        yield self.get_test(
            r'\b__noSuchMethod__\b',
            'The __noSuchMethod__ property has been deprecated.',
            'The __noSuchMethod__ property has been deprecated. See %s for '
            'more information.' % MDN_DOC
            % 'Web/JavaScript/Reference/Global_Objects/Object/noSuchMethod',
            log_function=self.err.warning,
            compat_type='warning')

        yield self.get_test(
            r'\bsendAsBinary\b',
            'The function sendAsBinary() in XMLHttpRequest has been removed.',
            'The function sendAsBinary() in XMLHttpRequest has been removed. '
            'You can use send() with a Blob instead. See %s for more '
            'information.' % BUGZILLA_BUG % 853162,
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\blightweightThemes\.(usedThemes|isThemeSelected)\b',
            'The preferences used to store theme selection have changed.',
            'The preferences used to store theme selection have changed. See '
            '%s for more information.' % (BUGZILLA_BUG % 1094821 + '#c39'),
            log_function=self.err.warning,
            compat_type='error')


@register_generator
class Gecko40RegexTests(CompatRegexTestHelper):
    """Regex tests for Gecko 40 updates."""

    VERSION = FX40_DEFINITION

    def tests(self):
        yield self.get_test(
            r'\b([gs]etKeywordForBookmark|getURIForKeyword)\b',
            'The old keywords API is deprecated.',
            'The old keywords API is deprecated. You should use '
            'PlacesUtils.keywords instead. See %s for more information.'
            % MDN_DOC % 'Mozilla/Tech/Places/Using_the_Places_keywords_API',
            log_function=self.err.warning,
            compat_type='warning')

        yield self.get_test(
            r'\b(fuelIApplication|extIApplication)\b',
            'The FUEL library is now deprecated.',
            'The FUEL library is now deprecated. You should use the add-ons '
            'SDK or Services.jsm. See %s for more information.'
            % MDN_DOC % 'Mozilla/Tech/Toolkit_API/FUEL',
            log_function=self.err.warning,
            compat_type='warning')

        yield self.get_test(
            r'\bresource://gre/modules/Dict.jsm\b',
            'The Dict.jsm module has been removed.',
            'The Dict.jsm module has been removed. You can use the native Map '
            'object instead. See %s for more information.'
            % MDN_DOC % 'Web/JavaScript/Reference/Global_Objects/Map',
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\bsessionstore-state-write\b',
            "The \"sessionstore-state-write\" notification has been removed.",
            "The \"sessionstore-state-write\" notification has been removed. "
            'See %s for more information.' % BUGZILLA_BUG % 1157235,
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\bnsISSLErrorListener\b',
            'The nsISSLErrorListener interface has been removed.',
            'The nsISSLErrorListener interface has been removed. See %s for '
            'more information.' % BUGZILLA_BUG % 844351,
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r"""require\(['"]sdk/widget['"]\)""",
            'The widget module has been removed.',
            'The widget module has been removed. You can use ActionButton or '
            'ToggleButton instead. See %s for more information.'
            % 'https://developer.mozilla.org/en-US/Add-ons/SDK/'
              'High-Level_APIs/widget',
            log_function=self.err.warning,
            compat_type='error')


@register_generator
class Gecko43RegexTests(CompatRegexTestHelper):
    """Regex tests for Firefox 43 updates."""

    VERSION = FX43_DEFINITION

    def tests(self):
        yield self.get_test(
            r'\bmozFetchAsStream\b',
            'The mozFetchAsStream function has been removed.',
            'The mozFetchAsStream function has been removed. '
            'You should use one of the standard canvas APIs instead. '
            'See %s for more information.' % MDN_DOC % 'Web/API/HTMLCanvasElement',
            log_function=self.err.warning,
            compat_type='error')


@register_generator
class Gecko44RegexTests(CompatRegexTestHelper):
    """Regex tests for Firefox 44 updates."""

    VERSION = FX44_DEFINITION

    def tests(self):
        yield self.get_test(
            r'resource://(?:gre)?/modules/devtools/',
            'The paths for all devtools JS modules have changed.',
            'The paths for all devtools JS modules have changed from '
            'resource://gre/modules/devtools/* to resource://devtools/*.'
            'See %s for more information.' % BUGZILLA_BUG % 1203159,
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\b__noSuchMethod__\b',
            '__noSuchMethod__ has been removed.',
            '__noSuchMethod__ has been removed. '
            'Proxies should be used instead, if necessary. See %s for '
            'more information.' % MDN_DOC
            % 'Web/JavaScript/Reference/Global_Objects/Proxy',
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\bgetAllStyleSheets\b',
            'The getAllStyleSheets function has been removed.',
            'The getAllStyleSheets function has been removed. '
            'You can use getBrowserStyleSheets instead. '
            'See %s for more information. ' % BUGZILLA_BUG % 1141041,
            log_function=self.err.warning,
            compat_type='error')


@register_generator
class Gecko45RegexTests(CompatRegexTestHelper):
    """Regex tests for Firefox 45 updates."""

    VERSION = FX45_DEFINITION

    def tests(self):
        yield self.get_test(
            r'\bnsIURIChecker\b',
            'The nsIURIChecker interface has been removed',
            'The nsIURIChecker interface has been removed. '
            'See %s for more information.' % BUGZILLA_BUG % 1222829,
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\b(gProxyFavIcon|\"page-proxy-favicon\")\b',
            (
                'The site identity interface has changed, '
                'which means gProxyFavIcon and "page-proxy-favicon" '
                'are no longer valid.'
            ),
            (
                'The site identity interface has changed, '
                'which means gProxyFavIcon and "page-proxy-favicon" '
                'are no longer valid. See %s for more information'
                % BUGZILLA_BUG % 1206244
            ),
            log_function=self.err.warning,
            compat_type='error')

        yield self.get_test(
            r'\bremoveAllPages\b',
            'The removeAllPages function is now deprecated.',
            'The removeAllPages function is now deprecated. '
            'You should use PlacesUtils.history.clear(); instead. '
            'See %s for more information.' % BUGZILLA_BUG % 1124185,
            log_function=self.err.warning,
            compat_type='warning')


@register_generator
class Gecko46RegexTests(CompatRegexTestHelper):
    """Regex tests for Firefox 46 updates."""

    VERSION = FX46_DEFINITION

    def tests(self):
        yield self.get_test(
            r'\b(mTabListeners|mTabFilters)\b',
            (
                'mTabListeners and mTabFilters are now Map objects and have '
                'been renamed to _tabListeners and _tabFilters, respectively.'
            ),
            (
                'mTabListeners and mTabFilters are now Map objects and have '
                'been renamed to _tabListeners and _tabFilters, respectively. '
                'See %s for more information' % BUGZILLA_BUG % 1238685
            ),
            log_function=self.err.warning,
            compat_type='error')


#############################
#  Thunderbird Regex Tests  #
#############################


@register_generator
class Thunderbird29RegexTests(CompatRegexTestHelper):

    VERSION = TB29_DEFINITION

    def tests(self):
        """String changes for Thunderbird 29 update."""
        patterns = {r'update\.checkingAddonCompat': 707489,
                    r'columnChooser\.tooltip': 881073,
                    r'threadColumn\.tooltip': 881073,
                    r'fromColumn\.tooltip': 881073,
                    r'recipientColumn\.tooltip': 881073,
                    r'attachmentColumn\.tooltip': 881073,
                    r'subjectColumn\.tooltip': 881073,
                    r'dateColumn\.tooltip': 881073,
                    r'priorityColumn\.tooltip': 881073,
                    r'tagsColumn\.tooltip': 881073,
                    r'accountColumn\.tooltip': 881073,
                    r'statusColumn\.tooltip': 881073,
                    r'sizeColumn\.tooltip': 881073,
                    r'junkStatusColumn\.tooltip': 881073,
                    r'unreadColumn\.tooltip': 881073,
                    r'totalColumn\.tooltip': 881073,
                    r'readColumn\.tooltip': 881073,
                    r'receivedColumn\.tooltip': 881073,
                    r'flagColumn\.tooltip': 881073,
                    r'starredColumn\.tooltip': 881073,
                    r'locationColumn\.tooltip': 881073,
                    r'idColumn\.tooltip': 881073,
                    r'phishingOptionDisableDetection\.label': 926473,
                    r'phishingOptionDisableDetection\.accesskey': 926473,
                    r'contextEditAsNew\.label': 956481,
                    r'contextEditAsNew\.accesskey': 956481,
                    r'EditContact\.label': 956481,
                    r'EditContact\.accesskey': 956481,
                    r'choosethisnewsserver\.label': 878805,
                    r'moveHereMenu\.label': 878805,
                    r'moveHereMenu\.accesskey': 878805,
                    r'newfolderchoosethis\.label': 878805,
                    r'thisFolder\.label': 878805,
                    r'thisFolder\.accesskey': 878805,
                    r'fileHereMenu\.label': 878805,
                    r'fileHereMenu\.accesskey': 878805,
                    r'copyHereMenu\.label': 878805,
                    r'copyHereMenu\.accesskey': 878805,
                    r'autoCheck\.label': 958850,
                    r'enableAppUpdate\.label': 958850,
                    r'enableAppUpdate\.accesskey': 958850,
                    r'enableAddonsUpdate\.label': 958850,
                    r'enableAddonsUpdate\.accesskey': 958850,
                    r'whenUpdatesFound\.label': 958850,
                    r'modeAskMe\.label': 958850,
                    r'modeAskMe\.accesskey': 958850,
                    r'modeAutomatic\.label': 958850,
                    r'modeAutomatic\.accesskey': 958850,
                    r'modeAutoAddonWarn\.label': 958850,
                    r'modeAutoAddonWarn\.accesskey': 958850,
                    r'showUpdates\.label': 958850,
                    r'showUpdates\.accesskey': 958850,
                    r'update\.checkInsideButton\.label': 707489,
                    r'update\.checkInsideButton\.accesskey': 707489,
                    r'update\.resumeButton\.label': 707489,
                    r'update\.resumeButton\.accesskey': 707489,
                    r'update\.openUpdateUI\.applyButton\.label': 707489,
                    r'update\.openUpdateUI\.applyButton\.accesskey': 707489,
                    r'update\.restart\.updateButton\.label': 707489,
                    r'update\.restart\.updateButton\.accesskey': 707489,
                    r'update\.restart\.restartButton\.label': 707489,
                    r'update\.restart\.restartButton\.accesskey': 707489,
                    r'update\.openUpdateUI\.upgradeButton\.label': 707489,
                    r'update\.openUpdateUI\.upgradeButton\.accesskey': 707489,
                    r'update\.restart\.upgradeButton\.label': 707489,
                    r'update\.restart\.upgradeButton\.accesskey': 707489,
                    r'command\.invite': 920801,
                    r'ctcp\.ping': 957918,
                    r'vkontakte\.usernameHint': 957918,
                    r'dateformat': 544315,}

        for pattern, bug in patterns.iteritems():
            yield self.get_test_bug(
                    bug, pattern,
                    'Removed labels in use.',
                    'Some string matched the pattern `%s`, which has been '
                    'flagged as having been removed or renamed '
                    'in Thunderbird 29.' % pattern,
                    compat_type='error')

@register_generator
class Thunderbird30RegexTests(CompatRegexTestHelper):

    VERSION = TB30_DEFINITION

    def tests(self):
        """String changes for Thunderbird 30 update."""
        patterns = {r'log\.lastWeek': 863226,
                    r'log\.twoWeeksAgo': 863226,
                    r'filemessageschoosethis\.label': 964425,
                    r'recentfolders\.label': 964425,
                    r'protocolNotFound\.title': 973368,
                    r'protocolNotFound\.longDesc': 973368,
                    r'quickFilterBar\.barLabel\.label': 592248,
                    r'updateOthers\.label': 978563,
                    r'enableAddonsUpdate3\.label': 978563,
                    r'enableAddonsUpdate3\.accesskey': 978563,
                    r'bounceSystemDockIcon\.label': 601263,
                    r'bounceSystemDockIcon\.accesskey': 601263,}

        for pattern, bug in patterns.iteritems():
            yield self.get_test_bug(
                    bug, pattern,
                    'Removed labels in use.',
                    'Some string matched the pattern `%s`, which has been '
                    'flagged as having been removed or renamed '
                    'in Thunderbird 30.' % pattern,
                    compat_type='error')

@register_generator
class Thunderbird31RegexTests(CompatRegexTestHelper):

    VERSION = TB31_DEFINITION

    def tests(self):
        """String changes for Thunderbird 31 update."""
        patterns = {r'youSendItMgmt\.viewSettings': 894306,
                    r'youSendItSettings\.username': 894306,
                    r'youSendItMgmt\.needAnAccount': 894306,
                    r'youSendItMgmt\.learnMore': 894306,
                    r'preferencesCmd\.label': 992643,
                    r'preferencesCmd\.accesskey': 992643,
                    r'proxy\.label': 992643,
                    r'proxy\.accesskey': 992643,
                    r'folderPropsCmd\.label': 992643,
                    r'folderPropsFolderCmd\.label': 992643,
                    r'folderPropsNewsgroupCmd\.label': 992643,
                    r'filtersCmd\.label': 992643,
                    r'filtersCmd\.accesskey': 992643,
                    r'accountManagerCmd\.accesskey': 992643,
                    r'accountManagerCmdUnix\.accesskey': 992643,
                    r'accountManagerCmd\.label': 992643,
                    r'accountManagerCmd\.accesskey': 992643,
                    r'accountManagerCmdUnix\.accesskey': 992643,
                    r'preferencesCmd\.label': 992643,
                    r'preferencesCmd\.accesskey': 992643,
                    r'preferencesCmdUnix\.label': 992643,
                    r'preferencesCmdUnix\.accesskey': 992643,
                    r'findCmd\.label': 530629,
                    r'findCmd\.key': 530629,
                    r'findCmd\.accesskey': 530629,
                    r'ubuntuOneMgmt\.viewSettings': 991220,
                    r'UbuntuOneSettings\.emailAddress': 991220,
                    r'UbuntuOneSettings\.needAnAccount': 991220,
                    r'UbuntuOneSettings\.learnMore': 991220,
                    r'propertiesCmd\.label': 992643,
                    r'propertiesCmd\.accesskey': 992643,
                    r'settingsOfflineCmd\.label': 992643,
                    r'settingsOfflineCmd\.accesskey': 992643,
                    r'folderContextProperties\.label': 992643,
                    r'folderContextProperties\.accesskey': 992643,
                    r'folderContextSettings\.label': 992643,
                    r'folderContextSettings\.accesskey': 992643,
                    r'itemCookies\.label': 953426,
                    r'cookies\.intro': 953426,
                    r'doNotTrack\.label': 953426,
                    r'doNotTrack\.accesskey': 953426,
                    r'allowRemoteContent1\.label': 457296,
                    r'allowRemoteContent1\.accesskey': 457296,
                    r'allowRemoteContent1\.tooltip': 457296,
                    r'remoteContentOptionAllowForAddress\.label': 457296,
                    r'remoteContentOptionAllowForAddress\.accesskey': 457296,
                    r'\b12504\b': 802266,
                    r'\b12505\b': 802266,
                    r'\b12507\b': 802266,
                    r'\b12522\b': 802266,
                    r'\b12508\b': 802266,
                    r'\b12509\b': 802266,
                    r'\b12521\b': 802266,
                    r'\b12523\b': 802266,
                    r'\b12533\b': 802266,
                    r'\b12534\b': 802266,
                    r'\b12535\b': 802266,
                    r'\b12536\b': 802266,
                    r'\b12537\b': 802266,
                    r'\b12538\b': 802266,
                    r'\b12539\b': 802266,
                    r'\b12540\b': 802266,
                    r'\b12541\b': 802266,
                    r'\b12550\b': 802266,
                    r'\b12551\b': 802266,
                    r'\b12556\b': 802266,
                    r'\b12557\b': 802266,
                    r'\b12558\b': 802266,
                    r'\b12559\b': 802266,
                    r'\b12562\b': 802266,
                    r'\b12566\b': 802266,
                    r'tooltip\.idleTime': 987577,
                    r'receivingMsgs': 86233,
                    r'hostContacted': 86233,
                    r'noMessages': 86233,
                    r'receivedMessages': 86233,
                    r'mailnews\.reply_header_authorwrote': 995797,
                    r'mailnews\.reply_header_ondate': 995797,}

        for pattern, bug in patterns.iteritems():
            yield self.get_test_bug(
                    bug, pattern,
                    'Removed labels in use.',
                    'Some string matched the pattern `%s`, which has been '
                    'flagged as having been removed or renamed '
                    'in Thunderbird 31.' % pattern,
                    compat_type='error')


class RegexTest(object):
    """
    Compiles a list of regular expressions (or iterables of literal strings)
    into a singular regular expression, and emits warnings for each match.
    """

    def __init__(self, regexps):
        key = 'test_%d'

        self.tests = {key % i: data
                      for i, (regexp, data) in enumerate(regexps)}

        # Used in tests.
        self.regex_source = '|'.join(
            r'(?P<%s>%s)' % (key % i, self.process_key(regexp))
            for i, (regexp, data) in enumerate(regexps))

        self.regex = re.compile(self.regex_source)

    def process_key(self, key):
        """Processes a key into a regular expression. Currently turns an
        enumerable value into a regexp which matches a string which is
        exactly equal to any included value."""

        if isinstance(key, basestring):
            return key

        return r'^(?:%s)$' % '|'.join(map(re.escape, key))

    def test(self, string, traverser, properties=None, wrapper=None):
        for match in self.regex.finditer(string):
            for key, val in match.groupdict().iteritems():
                if val is not None and key in self.tests:
                    kw = {'err_id': ('testcases_regex', 'string', 'generic')}
                    kw.update(self.tests[key])
                    if properties:
                        kw.update(properties)

                    msg = traverser.warning(**kw)
                    if wrapper:
                        wrapper.value.messages.append(msg)


PROFILE_FILENAMES = (
    'SiteSecurityServiceState.txt',
    'addons.json',
    'addons.sqlite',
    'blocklist.xml',
    'cert8.db',
    'compatibility.ini',
    'compreg.dat',
    'content-prefs.sqlite',
    'cookies.sqlite',
    'directoryLinks.json',
    'extensions.ini',
    'extensions.json',
    'extensions.sqlite',
    'formhistory.sqlite',
    'healthreport.sqlite',
    'httpDataUsage.dat',
    'key3.db',
    'localstore.rdf',
    'logins.json',
    'permissions.sqlite',
    'places.sqlite',
    'places.sqlite-shm',
    'places.sqlite-wal',
    'pluginreg.dat',
    'prefs.js',
    'safebrowsing/*',
    'search-metadata.json',
    'search.json',
    'search.sqlite',
    'searchplugins/*',
    'secmod.db',
    'sessionCheckpoints.json',
    'sessionstore.js',
    'signons.sqlite',
    'startupCache/*',
    'urlclassifier.pset',
    'urlclassifier3.sqlite',
    'urlclassifierkey3.txt',
    'user.js',
    'webappsstore.sqlite',
    'xpti.dat',
    'xulstore.json')
# These tests have proved too generic, and will need fine tuning:
#   "healthreport/*",
#   "storage/*",
#   "webapps/*",


def munge_filename(name):
    # Changes filenames ending with `/*` into a regular expression which
    # will also match sub-paths across platforms. Escapes regex metacharacters
    # in other strings.
    if name.endswith('/*'):
        return r'%s(?:[/\\].*)?' % re.escape(name[:-2])
    return re.escape(name)

PROFILE_REGEX = r'(?:^|[/\\])(?:%s)$' % '|'.join(map(munge_filename,
                                                     PROFILE_FILENAMES))

STRING_REGEXPS = (
    # Unsafe files in the profile directory.
    (PROFILE_REGEX, {
        'err_id': ('testcases_regex', 'string', 'profile_filenames'),
        'warning': 'Reference to critical user profile data',
        'description': 'Critical files in the user profile should not be '
                       'directly accessed by add-ons. In many cases, an '
                       'equivalent API is available and should be used '
                       'instead.',
        'signing_help': 'Please avoid touching files in the user profile '
                        'which do not belong to your add-on. If the effects '
                        'that you are trying to achieve cannot be replicated '
                        'with a built-in API, we strongly encourage you to '
                        'remove this functionality.',
        'signing_severity': 'low'}),

    # The names of potentially dangerous category names for the
    # category manager.
    (DANGEROUS_CATEGORIES, DANGEROUS_CATEGORY_WARNING),

    # References to the obsolete extension manager API.
    (r'@mozilla\.org/extensions/manager;1|'
     r'em-action-requested',
     {'warning': 'Obsolete Extension Manager API',
      'description': 'The old Extension Manager API is not available in any '
                     'remotely modern version of Firefox and should not be '
                     'referenced in any code.'}),

    # References to the Marionette service.
    (r'@mozilla\.org/marionette;1', MARIONETTE_MESSAGE),
    (r'\{786a1369-dca5-4adc-8486-33d23c88010a\}', MARIONETTE_MESSAGE),
)

PREFERENCE_ERROR_ID = 'testcases_regex', 'string', 'preference'

PREF_REGEXPS = (
    tuple(
        (pattern,
         {'err_id': PREFERENCE_ERROR_ID,
          'warning': 'Potentially unsafe preference branch referenced',
          'description': 'Extensions should not alter preferences '
                         'matching /%s/.' % pattern})
        for pattern in BANNED_PREF_REGEXPS) +
    tuple(
        ('^%s' % re.escape(branch),
         merge_description(
             {'err_id': PREFERENCE_ERROR_ID,
              'warning': 'Potentially unsafe preference branch referenced'},
             reason or ('Extensions should not alter preferences in '
                        'the `%s` preference branch' % branch)))
        for branch, reason in BANNED_PREF_BRANCHES))

COMPAT_PREF_REGEXPS = [
    (r'\bbrowser.newtab.url\b', {
        'message': 'The browser.newtab.url preference has been removed.',
        'warning': 'The browser.newtab.url preference has been removed. You '
                   'can use the override function in NewTabURL.jsm instead. '
                   'See %s for more information.' %
                       'http://mxr.mozilla.org/mozilla-beta/source/browser/'
                       'modules/NewTabURL.jsm',
        'tier': 5,
        'for_appversions': FX41_DEFINITION,
        'compatibility_type': 'error',
    }),
]

# For tests in literal strings, add help text suggesting passing the
# preference directly to reference getter functions.
PREF_STRING_HELP = (
    'If you are reading, but not writing, this preference, please consider '
    'passing a string literal directly to `Preferences.get()` or '
    '`nsIPrefBranch.get*Pref`.')


def maybe_tuple(value):
    """Return `value` as a tuple. If it is already a tuple, return it
    unchanged. Otherwise return a 1-element tuple containing `value`."""

    if isinstance(value, tuple):
        return value
    return (value,)


def add_pref_help(desc):
    desc = desc.copy()
    for key in 'description', 'signing_help':
        if key in desc:
            desc[key] = maybe_tuple(desc[key]) + maybe_tuple(PREF_STRING_HELP)

    return desc

STRING_REGEXPS += tuple((pattern, add_pref_help(desc))
                        for pattern, desc in PREF_REGEXPS)

# The following patterns should only be flagged in strings we're certain are
# being passed to preference setter functions, so add them after appending
# the others to the literal string tests.
PREF_REGEXPS += (
    (r'.*password.*',
     {'err_id': PREFERENCE_ERROR_ID,
      'warning': 'Passwords should not be stored in preferences',
      'description': 'Storing passwords in preferences is insecure. '
                     'The Login Manager should be used instead.'}),
)

validate_string = RegexTest(STRING_REGEXPS).test
validate_pref = RegexTest(PREF_REGEXPS).test
validate_compat_pref = RegexTest(COMPAT_PREF_REGEXPS).test



from validator import decorator


MANIFEST_URI = 'https://developer.mozilla.org/en/XUL_Tutorial/Manifest_Files'

DANGEROUS_CATEGORIES = ('JavaScript-global-constructor',
                        'JavaScript-global-constructor-prototype-alias',
                        'JavaScript-global-property',
                        'JavaScript-global-privileged-property',
                        'JavaScript-global-static-nameset',
                        'JavaScript-global-dynamic-nameset',
                        'JavaScript-DOM-class',
                        'JavaScript-DOM-interface')

DANGEROUS_CATEGORY_WARNING = {
    'err_id': ('testcases_chromemanifest', 'test_resourcemodules',
               'resource_modules'),
    'warning': 'Potentially dangerous category entry',
    'description': 'Add-ons defining global properties via category '
                   'entries require careful review by an administrative '
                   'reviewer.',
    'signing_help': (
        'Given the potential security risks of exposing APIs to unprivileged '
        'code, extensions which use these APIs must undergo manual code '
        'review for at least one submission. If you are not using these APIs '
        'to interact with content code, please consider alternatives, such as '
        'JavaScript modules (http://mzl.la/1HMH2m9), CommonJS modules '
        '(http://mzl.la/1JBMjuU, http://mzl.la/1OBaE8u), the observer '
        'service (http://mzl.la/1MLqWdJ), or window listeners which install '
        'global properties on privileged windows.'),
    'signing_severity': 'medium',
    'editors_only': True}


@decorator.register_test(tier=2, simple=True)
def test_categories(err):
    """Test for categories in the chrome.manifest file."""

    chrome = err.get_resource('chrome.manifest')
    if not chrome:
        return

    for triple in chrome.triples:
        if (triple['subject'] == 'category' and
            (triple['predicate'] in DANGEROUS_CATEGORIES or
             (triple['predicate'] == 'JavaScript' and
              (triple['object'].startswith('global ') or
               triple['object'].startswith('DOM '))))):
            err.warning(filename=triple['filename'],
                        line=triple['line'],
                        context=triple['context'],
                        **DANGEROUS_CATEGORY_WARNING)


@decorator.register_test(tier=2, simple=True)
def test_resourcemodules(err):
    """Flag instances of 'resource modules' in chrome.manifest."""

    chrome = err.get_resource('chrome.manifest')
    if not chrome:
        return

    for triple in chrome.triples:
        if (triple['subject'] == 'resource' and
                triple['predicate'].startswith('modules')):
            err.error(
                err_id=('testcases_chromemanifest', 'test_resourcemodules',
                        'resource_modules'),
                error="Resources should not be packages in the 'modules' "
                      'namespace',
                description='There should not be resources in the '
                            'chrome.manifest file that are listed as '
                            "'resource modules'.",
                filename=triple['filename'],
                line=triple['line'],
                context=triple['context'])


@decorator.register_test(tier=3, simple=True)
def test_content_instructions(err):
    """Flag content instructions which are not valid."""

    chrome = err.get_resource('chrome.manifest')
    if not chrome:
        return

    banned_namespaces = {
            'godlikea': "The 'godlikea' namespace is generated from a "
                        'template and should be replaced with something '
                        'unique to your add-on to avoid name conflicts.'}

    for triple in chrome.get_triples(subject='content'):
        if not triple['predicate'] or not triple['object']:
            err.warning(
                err_id=('testcases_chromemanifest',
                        'test_content_instructions', 'missing_triplicates'),
                warning='`content` instruction missing information',
                description='All content instructions must have a package '
                            'name and a URI to the files it describes.',
                filename=triple['filename'],
                line=triple['line'],
                context=triple['context'])
            continue

        if triple['predicate'] in banned_namespaces:
            err.error(
                err_id=('testcases_chromemanifest',
                        'test_content_instructions', 'godlikea'),
                error='Banned namespace in chrome.manifest',
                description=banned_namespaces[triple['predicate']],
                filename=triple['filename'],
                line=triple['line'],
                context=triple['context'])
        elif (triple['object'] != '' and
              not triple['object'].split()[0].endswith('/')):
            err.notice(
                err_id=('testcases_chromemanifest',
                        'test_content_instructions', 'trailing'),
                notice='Content instruction URIs must end with trailing slash',
                description='Chrome manifest content instructions must have a '
                            'trailing slash on their URI. For more '
                            'information, see %s.' % MANIFEST_URI,
                filename=triple['filename'],
                line=triple['line'],
                context=triple['context'])



import types
from StringIO import StringIO
from validator.contextgenerator import ContextGenerator

try:
    from html.parser import HTMLParser, HTMLParseError
except ImportError:
    from HTMLParser import HTMLParser, HTMLParseError


class DTDParser(object):
    'Parses and serializes DTD files. This is useful for L10n tests.'

    def __init__(self, dtd):
        """
        Creation of DTD parsers can be done based on a local file
        (provided as a string to the path), or directly (in memory as a
        StringIO object).
        """

        self.entities = {}
        self.items = []

        data = ''
        if isinstance(dtd, types.StringTypes):
            with open(dtd) as dtd_instance:
                data = dtd_instance.read()
        elif isinstance(dtd, file):
            data = dtd.read()
        elif isinstance(dtd, StringIO):
            data = dtd.getvalue()

        self._parse(data)

        # Create a context for the file
        self.context = ContextGenerator(data)

    def __len__(self):
        return len(self.entities)

    def _parse(self, data):
        'Parses the DTD data and stores it in an aggregate format.'

        parser = DTDXMLParser()
        # Feed the DTD file in line-by-line.
        for split_line in data.split('\n'):
            try:
                parser.feed(split_line + '\n')
            except HTMLParseError:
                parser = DTDXMLParser()
            else:
                if parser.out_buffer:
                    for name, value, line in parser.out_buffer:
                        self.entities[name] = value
                        self.items.append((name, value, line))
                    parser.clear_buffer()


class DTDXMLParser(HTMLParser):
    'Parses the individual XML entities in a DTD document.'

    def __init__(self):
        HTMLParser.__init__(self)
        self.out_buffer = []

    # Support for py2.7/3k
    def handle_comment(self, data):
        self.unknown_decl(data)

    def unknown_decl(self, decl):
        'Handles non-DOCTYPE SGML declarations in *ML documents.'

        decl = decl.strip()
        split_decl = decl.split()

        if len(split_decl) < 3 or split_decl[0] != 'ENTITY':
            # Interestingly enough, it legitimately IS an unknown
            # declaration. Funny thing, you know?
            return

        self.out_buffer.append((split_decl[1],
                                split_decl[2].strip('\'"'),
                                self.getpos()[0]))  # Pos 0 is the line no.

    def clear_buffer(self):
        'Clears the return buffer.'
        self.out_buffer = []



from helper import CompatTestCase
from validator.compat import FX38_DEFINITION


class TestFX38Compat(CompatTestCase):
    """Test that compatibility tests for Gecko 38 are properly executed."""

    VERSION = FX38_DEFINITION

    def test_mozIndexedDB(self):
        self.run_script_for_compat("""
            var db = mozIndexedDB.open('foo', 5);
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_indexedDB(self):
        self.run_script_for_compat("""
            var db = indexedDB.open('foo', 5);
        """)
        self.assert_silent()
        self.assert_compat_silent()

    def test_nsICompositionStringSynthesizer(self):
        self.run_script_for_compat("""
            var syn = Components.interfaces.nsICompositionStringSynthesizer;
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_sendCompositionEvent(self):
        self.run_script_for_compat("""
            var domWindowUtils = window
                .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                .getInterface(Components.interfaces.nsIDOMWindowUtils);
            domWindowUtils.sendCompositionEvent("compositionstart", "", "");
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_createCompositionStringSynthesizer(self):
        self.run_script_for_compat("""
            var domWindowUtils = window
                .QueryInterface(Components.interfaces.nsIInterfaceRequestor)
                .getInterface(Components.interfaces.nsIDOMWindowUtils);
            var compositionStringSynthesizer = domWindowUtils
                .createCompositionStringSynthesizer();
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_asyncFetch2(self):
        self.run_script_for_compat("""
            NetUtil.asyncFetch2(
                NetUtil.newURI(download.source.url),
                null,
                null,
                null,      // aLoadingNode
                Services.scriptSecurityManager.getSystemPrincipal(),
                null,      // aTriggeringPrincipal
                Ci.nsILoadInfo.SEC_NORMAL,
                Ci.nsIContentPolicy.TYPE_OTHER);
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_asyncFetch(self):
        self.run_script_for_compat("""
            NetUtil.asyncFetch({
                uri: "http://localhost:5555/test",
                loadUsingSystemPrincipal: true,
            });
        """)
        self.assert_silent()
        self.assert_compat_silent()

    def test_newChannel2(self):
        self.run_script_for_compat("""
            NetUtil.newChannel2(
                NetUtil.newURI(download.source.url),
                null,
                null,
                null,      // aLoadingNode
                Services.scriptSecurityManager.getSystemPrincipal(),
                null,      // aTriggeringPrincipal
                Ci.nsILoadInfo.SEC_NORMAL,
                Ci.nsIContentPolicy.TYPE_OTHER);
        """)
        self.assert_silent()
        self.assert_compat_warning(type_='warning')

    def test_newChannel(self):
        self.run_script_for_compat("""
            NetUtil.newChannel({
                uri: "data:text/plain,",
                loadUsingSystemPrincipal: true,
            });
        """)
        self.assert_silent()
        self.assert_compat_silent()

    def test_onProxyAvailable(self):
        self.run_script_for_compat("""
            onProxyAvailable(nsIChannel);
        """)
        self.assert_silent()
        self.assert_compat_error()

    def test_asyncResolve(self):
        self.run_script_for_compat("""
            asyncResolve(nsIChannel);
        """)
        self.assert_silent()
        self.assert_compat_error()
